import numpy as np

from qmllib.utils.alchemy import get_alchemy

from .fchl_kernel_functions import get_kernel_parameters
from .ffchl_module import (
    fget_ef_atomic_local_gradient_kernels_fchl,
    fget_ef_gaussian_process_kernels_fchl,
)


def get_atomic_local_electric_field_gradient_kernels(
    A,
    B,
    verbose=False,
    df=0.01,
    ef_scaling=0.01,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):
    """Calculates the Gaussian kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = \\exp \\big( -\\frac{\\|A_i - B_j\\|_2^2}{2\\sigma^2} \\big)`

    Where :math:`A_{i}` and :math:`B_{j}` are FCHL representation vectors.
    K is calculated analytically using an OpenMP parallel Fortran routine.
    Note, that this kernel will ONLY work with FCHL representations as input.

    :param A: Array of FCHL representation - shape=(N, maxsize, 5, maxneighbors).
    :type A: numpy array
    :param B: Array of FCHL representation - shape=(M, maxsize, 5, maxneighbors).
    :type B: numpy array

    :param two_body_scaling: Weight for 2-body terms.
    :type two_body_scaling: float
    :param three_body_scaling: Weight for 3-body terms.
    :type three_body_scaling: float

    :param two_body_width: Gaussian width for 2-body terms
    :type two_body_width: float
    :param three_body_width: Gaussian width for 3-body terms.
    :type three_body_width: float

    :param two_body_power: Powerlaw for :math:`r^{-n}` 2-body terms.
    :type two_body_power: float
    :param three_body_power: Powerlaw for Axilrod-Teller-Muto 3-body term
    :type three_body_power: float

    :param cut_start: The fraction of the cut-off radius at which cut-off damping start.
    :type cut_start: float
    :param cut_distance: Cut-off radius. (default=5 angstrom)
    :type cut_distance: float

    :param fourier_order: 3-body Fourier-expansion truncation order.
    :type fourier_order: integer
    :param alchemy: Type of alchemical interpolation ``"periodic-table"`` or ``"off"`` are possible options. Disabling alchemical interpolation can yield dramatic speedups.
    :type alchemy: string

    :param alchemy_period_width: Gaussian width along periods (columns) in the periodic table.
    :type alchemy_period_width: float
    :param alchemy_group_width: Gaussian width along groups (rows) in the periodic table.
    :type alchemy_group_width: float

    :return: Array of FCHL kernel matrices matrix - shape=(n_sigmas, N, M),
    :rtype: numpy array
    """

    atoms_max = A.shape[1]
    neighbors_max = A.shape[3]

    if not B.shape[1] == atoms_max:
        raise ValueError("Check FCHL representation sizes")
    if not B.shape[3] == neighbors_max:
        raise ValueError("Check FCHL representation sizes")

    nm1 = A.shape[0]
    nm2 = B.shape[0]

    N1 = np.zeros((nm1), dtype=np.int32)
    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    for a in range(nm2):
        N2[a] = len(np.where(B[a, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)
    neighbors2 = np.zeros((nm2, atoms_max), dtype=np.int32)

    for a, representation in enumerate(A):
        ni = N1[a]
        for i, x in enumerate(representation[:ni]):
            neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    for a, representation in enumerate(B):
        ni = N2[a]
        for i, x in enumerate(representation[:ni]):
            neighbors2[a, i] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, n_kernels = get_kernel_parameters(kernel, kernel_args)

    na1 = np.sum(N1)

    return fget_ef_atomic_local_gradient_kernels_fchl(
        A,
        B,
        verbose,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        na1,
        n_kernels,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        ef_scaling,
        df,
        kernel_idx,
        kernel_parameters,
    )


def get_gaussian_process_electric_field_kernels(
    A,
    B,
    verbose=False,
    fields=None,
    df=0.01,
    ef_scaling=0.01,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):
    """Calculates the Gaussian kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = \\exp \\big( -\\frac{\\|A_i - B_j\\|_2^2}{2\\sigma^2} \\big)`

    Where :math:`A_{i}` and :math:`B_{j}` are FCHL representation vectors.
    K is calculated analytically using an OpenMP parallel Fortran routine.
    Note, that this kernel will ONLY work with FCHL representations as input.

    :param A: Array of FCHL representation - shape=(N, maxsize, 5, maxneighbors).
    :type A: numpy array
    :param B: Array of FCHL representation - shape=(M, maxsize, 5, maxneighbors).
    :type B: numpy array

    :param two_body_scaling: Weight for 2-body terms.
    :type two_body_scaling: float
    :param three_body_scaling: Weight for 3-body terms.
    :type three_body_scaling: float

    :param two_body_width: Gaussian width for 2-body terms
    :type two_body_width: float
    :param three_body_width: Gaussian width for 3-body terms.
    :type three_body_width: float

    :param two_body_power: Powerlaw for :math:`r^{-n}` 2-body terms.
    :type two_body_power: float
    :param three_body_power: Powerlaw for Axilrod-Teller-Muto 3-body term
    :type three_body_power: float

    :param cut_start: The fraction of the cut-off radius at which cut-off damping start.
    :type cut_start: float
    :param cut_distance: Cut-off radius. (default=5 angstrom)
    :type cut_distance: float

    :param fourier_order: 3-body Fourier-expansion truncation order.
    :type fourier_order: integer
    :param alchemy: Type of alchemical interpolation ``"periodic-table"`` or ``"off"`` are possible options. Disabling alchemical interpolation can yield dramatic speedups.
    :type alchemy: string

    :param alchemy_period_width: Gaussian width along periods (columns) in the periodic table.
    :type alchemy_period_width: float
    :param alchemy_group_width: Gaussian width along groups (rows) in the periodic table.
    :type alchemy_group_width: float

    :return: Array of FCHL kernel matrices matrix - shape=(n_sigmas, N, M),
    :rtype: numpy array
    """

    atoms_max = A.shape[1]
    neighbors_max = A.shape[3]

    if not B.shape[1] == atoms_max:
        raise ValueError("Check FCHL representation sizes")
    if not B.shape[3] == neighbors_max:
        raise ValueError("Check FCHL representation sizes")

    nm1 = A.shape[0]
    nm2 = B.shape[0]

    N1 = np.zeros((nm1), dtype=np.int32)
    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    for a in range(nm2):
        N2[a] = len(np.where(B[a, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)
    neighbors2 = np.zeros((nm2, atoms_max), dtype=np.int32)

    for a, representation in enumerate(A):
        ni = N1[a]
        for i, x in enumerate(representation[:ni]):
            neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    for a, representation in enumerate(B):
        ni = N2[a]
        for i, x in enumerate(representation[:ni]):
            neighbors2[a, i] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, n_kernels = get_kernel_parameters(kernel, kernel_args)

    F1 = np.zeros((nm1, 3))
    F2 = np.zeros((nm2, 3))

    if fields is not None:

        F1 = np.array(fields[0])
        F2 = np.array(fields[1])

    return fget_ef_gaussian_process_kernels_fchl(
        A,
        B,
        verbose,
        F1,
        F2,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        n_kernels,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        ef_scaling,
        df,
        kernel_idx,
        kernel_parameters,
    )


def get_kernels_ef_field(
    A,
    B,
    fields=None,
    df=0.01,
    ef_scaling=0.01,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):

    raise NotImplementedError("Undefined fget_kernels_fchl_ef_field")

    # atoms_max = A.shape[1]
    # neighbors_max = A.shape[3]

    # assert B.shape[1] == atoms_max, "Check FCHL representation sizes! code = 2"
    # assert B.shape[3] == neighbors_max, "Check FCHL representation sizes! code = 3"

    # nm1 = A.shape[0]
    # nm2 = B.shape[0]

    # N1 = np.zeros((nm1), dtype=np.int32)
    # N2 = np.zeros((nm2), dtype=np.int32)

    # for a in range(nm1):
    #     N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    # for a in range(nm2):
    #     N2[a] = len(np.where(B[a, :, 1, 0] > 0.0001)[0])

    # neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)
    # neighbors2 = np.zeros((nm2, atoms_max), dtype=np.int32)

    # for a, representation in enumerate(A):
    #     ni = N1[a]
    #     for i, x in enumerate(representation[:ni]):
    #         neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    # for a, representation in enumerate(B):
    #     ni = N2[a]
    #     for i, x in enumerate(representation[:ni]):
    #         neighbors2[a, i] = len(np.where(x[0] < cut_distance)[0])

    # doalchemy, pd = get_alchemy(
    #     alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    # )

    # kernel_idx, kernel_parameters, n_kernels = get_kernel_parameters(kernel, kernel_args)

    # F2 = np.zeros((nm2, 3))

    # if fields is not None:

    #     F2 = np.array(fields)

    # na1 = np.sum(N1)

    # return fget_kernels_fchl_ef_field(
    #     A,
    #     B,
    #     F2,
    #     N1,
    #     N2,
    #     neighbors1,
    #     neighbors2,
    #     nm1,
    #     nm2,
    #     n_kernels,
    #     na1,
    #     three_body_width,
    #     two_body_width,
    #     cut_start,
    #     cut_distance,
    #     fourier_order,
    #     pd,
    #     two_body_scaling,
    #     three_body_scaling,
    #     doalchemy,
    #     two_body_power,
    #     three_body_power,
    #     ef_scaling,
    #     kernel_idx,
    #     kernel_parameters,
    # )


# Non-functional kernel for polarizabilites - unsure why it won't work

# def get_local_kernels_pol(A, B, df=1e-5, ef_scaling=1.0,
#         two_body_scaling=np.sqrt(8), three_body_scaling=1.6,
#         two_body_width=0.2, three_body_width=np.pi,
#         two_body_power=4.0, three_body_power=2.0,
#         cut_start=1.0, cut_distance=5.0,
#         fourier_order=1, alchemy="periodic-table",
#         alchemy_period_width=1.6, alchemy_group_width=1.6,
#         kernel="gaussian", kernel_args=None):
#     """ Calculates the Gaussian kernel matrix K, where :math:`K_{ij}`:

#             :math:`K_{ij} = \\exp \\big( -\\frac{\\|A_i - B_j\\|_2^2}{2\sigma^2} \\big)`

#         Where :math:`A_{i}` and :math:`B_{j}` are FCHL representation vectors.
#         K is calculated analytically using an OpenMP parallel Fortran routine.
#         Note, that this kernel will ONLY work with FCHL representations as input.

#         :param A: Array of FCHL representation - shape=(N, maxsize, 5, maxneighbors).
#         :type A: numpy array
#         :param B: Array of FCHL representation - shape=(M, maxsize, 5, maxneighbors).
#         :type B: numpy array

#         :param two_body_scaling: Weight for 2-body terms.
#         :type two_body_scaling: float
#         :param three_body_scaling: Weight for 3-body terms.
#         :type three_body_scaling: float

#         :param two_body_width: Gaussian width for 2-body terms
#         :type two_body_width: float
#         :param three_body_width: Gaussian width for 3-body terms.
#         :type three_body_width: float

#         :param two_body_power: Powerlaw for :math:`r^{-n}` 2-body terms.
#         :type two_body_power: float
#         :param three_body_power: Powerlaw for Axilrod-Teller-Muto 3-body term
#         :type three_body_power: float

#         :param cut_start: The fraction of the cut-off radius at which cut-off damping start.
#         :type cut_start: float
#         :param cut_distance: Cut-off radius. (default=5 angstrom)
#         :type cut_distance: float

#         :param fourier_order: 3-body Fourier-expansion truncation order.
#         :type fourier_order: integer
#         :param alchemy: Type of alchemical interpolation ``"periodic-table"`` or ``"off"`` are possible options. Disabling alchemical interpolation can yield dramatic speedups.
#         :type alchemy: string

#         :param alchemy_period_width: Gaussian width along periods (columns) in the periodic table.
#         :type alchemy_period_width: float
#         :param alchemy_group_width: Gaussian width along groups (rows) in the periodic table.
#         :type alchemy_group_width: float

#         :return: Array of FCHL kernel matrices matrix - shape=(n_sigmas, N, M),
#         :rtype: numpy array
#     """

#     atoms_max = A.shape[1]
#     neighbors_max = A.shape[3]

#     assert B.shape[1] == atoms_max, "Check FCHL representation sizes! code = 2"
#     assert B.shape[3] == neighbors_max, "Check FCHL representation sizes! code = 3"


#     nm1 = A.shape[0]
#     nm2 = B.shape[0]

#     N1 = np.zeros((nm1),dtype=np.int32)
#     N2 = np.zeros((nm2),dtype=np.int32)

#     for a in range(nm1):
#         N1[a] = len(np.where(A[a,:,1,0] > 0.0001)[0])

#     for a in range(nm2):
#         N2[a] = len(np.where(B[a,:,1,0] > 0.0001)[0])

#     neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)
#     neighbors2 = np.zeros((nm2, atoms_max), dtype=np.int32)

#     for a, representation in enumerate(A):
#         ni = N1[a]
#         for i, x in enumerate(representation[:ni]):
#             neighbors1[a,i] = len(np.where(x[0]< cut_distance)[0])

#     for a, representation in enumerate(B):
#         ni = N2[a]
#         for i, x in enumerate(representation[:ni]):
#             neighbors2[a,i] = len(np.where(x[0]< cut_distance)[0])

#     doalchemy, pd = get_alchemy(alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width)

#     kernel_idx, kernel_parameters, n_kernels = get_kernel_parameters(kernel, kernel_args)

#     na1 = np.sum(N1)

#     return fget_kernels_fchl_ef_2ndderiv(A, B, N1, N2, neighbors1, neighbors2, nm1, nm2, na1, n_kernels,
#                 three_body_width, two_body_width, cut_start, cut_distance, fourier_order, pd, two_body_scaling,
#                 three_body_scaling, doalchemy, two_body_power, three_body_power, ef_scaling, df, kernel_idx, kernel_parameters)



""" Compile script for Fortran """

import os
import subprocess
import sys
from pathlib import Path

import numpy as np

DEFAULT_FC = "gfortran"

f90_modules = {
    "representations/frepresentations": ["frepresentations.f90"],
    "representations/facsf": ["facsf.f90"],
    "representations/fslatm": ["fslatm.f90"],
    "representations/arad/farad_kernels": ["farad_kernels.f90"],
    "representations/fchl/ffchl_module": [
        "ffchl_kernel_types.f90",
        "ffchl_module.f90",
        "ffchl_module_ef.f90",
        "ffchl_kernels.f90",
        "ffchl_scalar_kernels.f90",
        "ffchl_kernels_ef.f90",
        "ffchl_force_kernels.f90",
    ],
    "solvers/fsolvers": ["fsolvers.f90"],
    "kernels/fdistance": ["fdistance.f90"],
    "kernels/fkernels": [
        "fkernels.f90",
        "fkpca.f90",
        "fkwasserstein.f90",
    ],
    "kernels/fgradient_kernels": ["fgradient_kernels.f90"],
    "utils/fsettings": ["fsettings.f90"],
}


def find_mkl():
    raise NotImplementedError()


def find_env() -> dict[str, str]:
    """Find compiler flag"""

    """
    For anaconda-like envs
        TODO Find MKL

    For brew,

        brew install llvm libomp
        brew install openblas lapack

        export LDFLAGS="-L/opt/homebrew/opt/lapack/lib"
        export CPPFLAGS="-I/opt/homebrew/opt/lapack/include"
        export LDFLAGS="-L/opt/homebrew/opt/libomp/lib"
        export CPPFLAGS="-I/opt/homebrew/opt/libomp/include"

    """

    fc = os.environ.get("FC", DEFAULT_FC)

    # TODO Check if FC is there, not not raise Error
    # TODO Check if lapack / blas is there, if not raise Error
    # TODO Check if omp is installed

    # TODO Find ifort flags, choose from FC
    # TODO Find mkl lib

    # TODO Check if darwin, check for brew paths

    # Default GNU flags
    compiler_flags = [
        "-O3",
        "-m64",
        "-march=native",
        "-fPIC",
        "-Wno-maybe-uninitialized",
        "-Wno-unused-function",
        "-Wno-cpp",
    ]
    compiler_openmp = [
        "-fopenmp",
    ]
    linker_flags = [
        "-lpthread",
        "-lm",
        "-ldl",
    ]
    linker_openmp = [
        "-lgomp",
    ]
    linker_math = [
        "-lblas",
        "-llapack",
        "-L/usr/lib/",
    ]

    # MacOS X specific flags
    if "darwin" in sys.platform:

        expected_omp_dir = Path("/opt/homebrew/opt/libomp/lib")

        if expected_omp_dir.is_dir():
            compiler_openmp = [
                "-fopenmp",
            ]
            linker_openmp = [
                f"-L{expected_omp_dir}",
                "-lomp",
            ]

        else:
            print(f"Expected OpenMP dir not found: {expected_omp_dir}, compiling without OpenMP")
            compiler_openmp = []
            linker_openmp = []

    # FreeBSD specific flags
    if "freebsd" in sys.platform:
        # Location of BLAS / Lapack for FreeBSD 14
        linker_math += ["-L/usr/local/lib/"]

    fflags = [] + compiler_flags + compiler_openmp
    ldflags = [] + linker_flags + linker_math + linker_openmp

    env = {"FFLAGS": " ".join(fflags), "LDFLAGS": " ".join(ldflags), "FC": fc}

    return env


def main():
    """Compile f90 in src/qmllib"""

    print(
        f"Using python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    )
    print(f"Using numpy {np.__version__}")

    # Find and set Fortran compiler, compiler flags and linker flags
    env = find_env()
    for key, value in env.items():
        print(f"export {key}='{value}'")
        os.environ[key] = value

    f2py = [sys.executable, "-m", "numpy.f2py"]

    meson_flags = [
        "--backend",
        "meson",
    ]

    for module_name, module_sources in f90_modules.items():

        path = Path(module_name)
        parent = path.parent
        stem = path.stem

        cwd = Path("src/qmllib") / parent
        cmd = f2py + ["-c"] + module_sources + ["-m", str(stem)] + meson_flags
        print(cwd, " ".join(cmd))

        proc = subprocess.run(cmd, cwd=cwd, capture_output=True, text=True)
        stdout = proc.stdout
        stderr = proc.stderr
        exitcode = proc.returncode

        if exitcode > 0:
            print(stderr)
            print()
            print(stdout)
            exit(exitcode)


if __name__ == "__main__":
    main()



from qmllib.representations.arad import generate_arad  # noqa:403
from qmllib.representations.fchl import (  # noqa:F403
    generate_fchl18,
    generate_fchl18_displaced,
    generate_fchl18_displaced_5point,
    generate_fchl18_electric_field,
)
from qmllib.representations.representations import (  # noqa:F403
    generate_acsf,
    generate_bob,
    generate_coulomb_matrix,
    generate_coulomb_matrix_atomic,
    generate_coulomb_matrix_eigenvalue,
    generate_fchl19,
    generate_slatm,
    get_slatm_mbtypes,
)



import copy
from typing import List, Optional, Union

import numpy as np
from numpy import ndarray


def generate_fchl18(
    nuclear_charges: ndarray,
    coordinates: Union[ndarray, List[List[float]]],
    max_size: int = 23,
    neighbors: int = 23,
    cut_distance: float = 5.0,
    cell: Optional[ndarray] = None,
) -> ndarray:
    """Generates a representation for the FCHL kernel module.

    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param coordinates: Input coordinates.
    :type coordinates: numpy array
    :param max_size: Max number of atoms in representation.
    :type max_size: integer
    :param neighbors: Max number of atoms within the cut-off around an atom. (For periodic systems)
    :type neighbors: integer
    :param cell: Unit cell vectors. The presence of this keyword argument will generate a periodic representation.
    :type cell: numpy array
    :param cut_distance: Spatial cut-off distance - must be the same as used in the kernel function call.
    :type cut_distance: float
    :return: FCHL representation, shape = (size,5,neighbors).
    :rtype: numpy array
    """

    size = max_size

    if cell is None:
        neighbors = size

    L = len(coordinates)
    coords = np.asarray(coordinates)
    ocupationList = np.asarray(nuclear_charges)
    M = np.zeros((size, 5, neighbors))

    if cell is not None:
        coords = np.dot(coords, cell)
        nExtend = (np.floor(cut_distance / np.linalg.norm(cell, 2, axis=0)) + 1).astype(int)

        for i in range(-nExtend[0], nExtend[0] + 1):
            for j in range(-nExtend[1], nExtend[1] + 1):
                for k in range(-nExtend[2], nExtend[2] + 1):
                    if i == -nExtend[0] and j == -nExtend[1] and k == -nExtend[2]:
                        coordsExt = coords + i * cell[0, :] + j * cell[1, :] + k * cell[2, :]
                        ocupationListExt = copy.copy(ocupationList)
                    else:
                        ocupationListExt = np.append(ocupationListExt, ocupationList)
                        coordsExt = np.append(
                            coordsExt,
                            coords + i * cell[0, :] + j * cell[1, :] + k * cell[2, :],
                            axis=0,
                        )
    else:
        coordsExt = copy.copy(coords)
        ocupationListExt = copy.copy(ocupationList)

    M[:, 0, :] = 1e100

    for i in range(L):
        cD = -coords[i] + coordsExt[:]
        ocExt = np.asarray(ocupationListExt)

        D1 = np.sqrt(np.sum(cD**2, axis=1))
        args = np.argsort(D1)
        D1 = D1[args]
        ocExt = np.asarray([ocExt[l] for l in args])
        cD = cD[args]

        args = np.where(D1 < cut_distance)[0]
        D1 = D1[args]
        ocExt = np.asarray([ocExt[l] for l in args])
        cD = cD[args]
        M[i, 0, : len(D1)] = D1
        M[i, 1, : len(D1)] = ocExt[:]
        M[i, 2:5, : len(D1)] = cD.T
    return M


def generate_fchl18_displaced(
    nuclear_charges, coordinates, max_size=23, neighbors=23, cut_distance=5.0, cell=None, dx=0.005
):
    """Generates displaced representations for the FCHL kernel module.

    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param coordinates: Input coordinates.
    :type coordinates: numpy array
    :param max_size: Max number of atoms in representation.
    :type max_size: integer
    :param neighbors: Max number of atoms within the cut-off around an atom. (For periodic systems)
    :type neighbors: integer
    :param cell: Unit cell vectors. The presence of this keyword argument will generate a periodic representation.
    :type cell: numpy array
    :param dx: Real-space displacement for numerical derivatives, in units of angstrom.
    :type dx: float
    :param cut_distance: Spatial cut-off distance - must be the same as used in the kernel function call.
    :type cut_distance: float
    :return: FCHL representation, shape = (size,5,neighbors).
    :rtype: numpy array
    """
    size = max_size
    if cell is None:
        neighbors = size
    reps = np.zeros((3, 2, size, size, 5, neighbors))

    compound_size = len(nuclear_charges)

    for xyz in range(3):

        for i in range(compound_size):
            for idisp, disp in enumerate([-dx, dx]):

                displaced_coordinates = copy.deepcopy(coordinates)
                displaced_coordinates[i, xyz] += disp

                rep = generate_fchl18(
                    nuclear_charges,
                    displaced_coordinates,
                    max_size=size,
                    neighbors=neighbors,
                    cut_distance=cut_distance,
                    cell=cell,
                )

                reps[xyz, idisp, i, :, :, :] = rep[:, :, :]

    return reps


def generate_fchl18_displaced_5point(
    nuclear_charges, coordinates, max_size=23, neighbors=23, cut_distance=5.0, cell=None, dx=0.005
):
    """Generates displaced representations for the FCHL kernel module, using a 5-point stencil.

    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param coordinates: Input coordinates.
    :type coordinates: numpy array
    :param max_size: Max number of atoms in representation.
    :type max_size: integer
    :param neighbors: Max number of atoms within the cut-off around an atom. (For periodic systems)
    :type neighbors: integer
    :param cell: Unit cell vectors. The presence of this keyword argument will generate a periodic representation.
    :type cell: numpy array
    :param dx: Real-space displacement for numerical derivatives, in units of angstrom.
    :type dx: float
    :param cut_distance: Spatial cut-off distance - must be the same as used in the kernel function call.
    :type cut_distance: float
    :return: FCHL representation, shape = (size,5,neighbors).
    :rtype: numpy array
    """
    size = max_size
    if cell is None:
        neighbors = size
    reps = np.zeros((3, 5, size, size, 5, neighbors))

    compound_size = len(nuclear_charges)

    for xyz in range(3):

        for i in range(compound_size):
            for idisp, disp in enumerate([-2 * dx, -dx, 0.0, dx, 2 * dx]):

                displaced_coordinates = copy.deepcopy(coordinates)
                displaced_coordinates[i, xyz] += disp

                rep = generate_fchl18(
                    nuclear_charges,
                    displaced_coordinates,
                    max_size=size,
                    neighbors=neighbors,
                    cut_distance=cut_distance,
                    cell=cell,
                )

                reps[xyz, idisp, i, :, :, :] = rep[:, :, :]

    return reps


def generate_fchl18_electric_field(
    nuclear_charges: ndarray,
    coordinates: ndarray,
    fictitious_charges: Union[ndarray, List[float]] = "gasteiger",
    max_size: int = 23,
    neighbors: int = 23,
    cut_distance: float = 5.0,
) -> ndarray:
    """Generates a representation for the FCHL kernel module, including fictitious partial charges.
        For use with fist-order electric field-dependent properties, such as dipole moments and IR intensity.

        Good choices are charges from e.g. a force field model or a QM charge model, e.g. Mulliken, etc.

    :param coordinates: Input coordinates.
    :type coordinates: numpy array
    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param partial_charges:  Either list of fictitious partial charges, or a string representing the charge model from which these are to be calculated (the latter requires Open Babel with Pybel installed). Default option is "gasteiger". Please see the OpenBabel documentation for list of supported charges models.
    :type partial_charges: numpy array or string.
    :param max_size: Max number of atoms in representation.
    :type max_size: integer
    :param neighbors: Max number of atoms within the cut-off around an atom. (For periodic systems)
    :type neighbors: integer
    :param cell: Unit cell vectors. The presence of this keyword argument will generate a periodic representation.
    :type cell: numpy array
    :param cut_distance: Spatial cut-off distance - must be the same as used in the kernel function call.
    :type cut_distance: float
    :return: FCHL representation, shape = (size,5,neighbors).
    :rtype: numpy array
    """

    partial_charges = None

    # If a list is given, assume these are the fictitious charges

    if isinstance(fictitious_charges, list) or isinstance(fictitious_charges, np.ndarray):

        if len(fictitious_charges) != len(nuclear_charges):
            raise ValueError("Error: incorrect length of fictitious charge list")

        partial_charges = fictitious_charges

    # # Otherwise, if a string is given, assume this is the name of a charge model
    # # in Open Babel//Pybel.
    # elif isinstance(fictitious_charges, (basestring,)):

    #     # Dirty hack for now.
    #     try:
    #         import openbabel
    #         import pybel
    #     except ImportError:
    #         print(
    #             "QML ERROR: Could not generate fictitious charges because OpenBabel/Pybel was not found."
    #         )
    #         exit()

    #     temp_xyz = "%i\n\n" % len(nuclear_charges)

    #     for i, nuc in enumerate(nuclear_charges):
    #         temp_xyz += "%s %f %f %f\n" % (
    #             ELEMENT_NAME[nuc],
    #             coordinates[i][0],
    #             coordinates[i][1],
    #             coordinates[i][2],
    #         )

    #     mol = pybel.readstring("xyz", temp_xyz)

    #     this_charge_model = openbabel.OBChargeModel.FindType(fictitious_charges)
    #     this_charge_model.ComputeCharges(mol.OBMol)

    #     partial_charges = [atom.partialcharge for atom in mol]

    else:
        raise ValueError("Missing fictitious charges")

    size = max_size
    neighbors = size

    L = len(coordinates)
    coords = np.asarray(coordinates)
    ocupationList = np.asarray(nuclear_charges)
    partial_charges = np.asarray(partial_charges)
    M = np.zeros((size, 6, neighbors))

    coordsExt = copy.copy(coords)
    partialExt = copy.copy(partial_charges)
    ocupationListExt = copy.copy(ocupationList)

    M[:, 0, :] = 1e100

    for i in range(L):
        cD = -coords[i] + coordsExt[:]

        ocExt = np.asarray(ocupationListExt)
        qExt = np.asarray(partialExt)

        D1 = np.sqrt(np.sum(cD**2, axis=1))
        args = np.argsort(D1)
        D1 = D1[args]

        ocExt = np.asarray([ocExt[l] for l in args])
        qExt = np.asarray([qExt[l] for l in args])

        cD = cD[args]

        args = np.where(D1 < cut_distance)[0]
        D1 = D1[args]
        ocExt = np.asarray([ocExt[l] for l in args])
        qExt = np.asarray([qExt[l] for l in args])

        cD = cD[args]
        M[i, 0, : len(D1)] = D1
        M[i, 1, : len(D1)] = ocExt[:]
        M[i, 2:5, : len(D1)] = cD.T
        M[i, 5, : len(D1)] = qExt[:]

    return M



import itertools
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from numpy import int64, ndarray

from qmllib.constants.periodic_table import NUCLEAR_CHARGE

from .facsf import (
    fgenerate_acsf,
    fgenerate_acsf_and_gradients,
    fgenerate_fchl_acsf,
    fgenerate_fchl_acsf_and_gradients,
)
from .frepresentations import (
    fgenerate_atomic_coulomb_matrix,
    fgenerate_bob,
    fgenerate_coulomb_matrix,
    fgenerate_eigenvalue_coulomb_matrix,
    fgenerate_local_coulomb_matrix,
    fgenerate_unsorted_coulomb_matrix,
)
from .slatm import get_boa, get_sbop, get_sbot


def vector_to_matrix(v):
    """Converts a representation from 1D vector to 2D square matrix.
    :param v: 1D input representation.
    :type v: numpy array
    :return: Square matrix representation.
    :rtype: numpy array
    """

    if not (np.sqrt(8 * v.shape[0] + 1) == int(np.sqrt(8 * v.shape[0] + 1))):
        raise ValueError("Can not make a square matrix.")

    n = v.shape[0]
    l = (-1 + int(np.sqrt(8 * n + 1))) // 2
    M = np.empty((l, l))

    index = 0
    for i in range(l):
        for j in range(l):
            if j > i:
                continue

            M[i, j] = v[index]
            M[j, i] = M[i, j]

            index += 1
    return M


def generate_coulomb_matrix(
    nuclear_charges: ndarray, coordinates: ndarray, size: int = 23, sorting: str = "row-norm"
) -> ndarray:
    """ Creates a Coulomb Matrix representation of a molecule.
        Sorting of the elements can either be done by ``sorting="row-norm"`` or ``sorting="unsorted"``.
        A matrix :math:`M` is constructed with elements

        .. math::

            M_{ij} =
              \\begin{cases}
                 \\tfrac{1}{2} Z_{i}^{2.4} & \\text{if } i = j \\\\
                 \\frac{Z_{i}Z_{j}}{\\| {\\bf R}_{i} - {\\bf R}_{j}\\|}       & \\text{if } i \\neq j
              \\end{cases},

        where :math:`i` and :math:`j` are atom indices, :math:`Z` is nuclear charge and
        :math:`\\bf R` is the coordinate in euclidean space.
        If ``sorting = 'row-norm'``, the atom indices are reordered such that

            :math:`\\sum_j M_{1j}^2 \\geq \\sum_j M_{2j}^2 \\geq ... \\geq \\sum_j M_{nj}^2`

        The upper triangular of M, including the diagonal, is concatenated to a 1D
        vector representation.

        If ``sorting = 'unsorted``, the elements are sorted in the same order as the input coordinates
        and nuclear charges.

        The representation is calculated using an OpenMP parallel Fortran routine.

        :param nuclear_charges: Nuclear charges of the atoms in the molecule
        :type nuclear_charges: numpy array
        :param coordinates: 3D Coordinates of the atoms in the molecule
        :type coordinates: numpy array
        :param size: The size of the largest molecule supported by the representation
        :type size: integer
        :param sorting: How the atom indices are sorted ('row-norm', 'unsorted')
        :type sorting: string

        :return: 1D representation - shape (size(size+1)/2,)
        :rtype: numpy array
    """

    if sorting == "row-norm":
        return fgenerate_coulomb_matrix(nuclear_charges, coordinates, size)

    elif sorting == "unsorted":
        return fgenerate_unsorted_coulomb_matrix(nuclear_charges, coordinates, size)

    else:
        raise ValueError("Unknown sorting scheme requested")


def generate_coulomb_matrix_atomic(
    nuclear_charges: ndarray,
    coordinates: ndarray,
    size: int = 23,
    sorting: str = "distance",
    central_cutoff: float = 1e6,
    central_decay: Union[float, int] = -1,
    interaction_cutoff: float = 1e6,
    interaction_decay: Union[float, int] = -1,
    indices: Optional[List[int]] = None,
) -> ndarray:
    """ Creates a Coulomb Matrix representation of the local environment of a central atom.
        For each central atom :math:`k`, a matrix :math:`M` is constructed with elements

        .. math::

            M_{ij}(k) =
              \\begin{cases}
                 \\tfrac{1}{2} Z_{i}^{2.4} \\cdot f_{ik}^2 & \\text{if } i = j \\\\
                 \\frac{Z_{i}Z_{j}}{\\| {\\bf R}_{i} - {\\bf R}_{j}\\|} \\cdot f_{ik}f_{jk}f_{ij} & \\text{if } i \\neq j
              \\end{cases},

        where :math:`i`, :math:`j` and :math:`k` are atom indices, :math:`Z` is nuclear charge and
        :math:`\\bf R` is the coordinate in euclidean space.

        :math:`f_{ij}` is a function that masks long range effects:

        .. math::

            f_{ij} =
              \\begin{cases}
                 1 & \\text{if } \\|{\\bf R}_{i} - {\\bf R}_{j} \\| \\leq r - \\Delta r \\\\
                 \\tfrac{1}{2} \\big(1 + \\cos\\big(\\pi \\tfrac{\\|{\\bf R}_{i} - {\\bf R}_{j} \\|
                    - r + \\Delta r}{\\Delta r} \\big)\\big)
                    & \\text{if } r - \\Delta r < \\|{\\bf R}_{i} - {\\bf R}_{j} \\| \\leq r - \\Delta r \\\\
                 0 & \\text{if } \\|{\\bf R}_{i} - {\\bf R}_{j} \\| > r
              \\end{cases},

        where the parameters ``central_cutoff`` and ``central_decay`` corresponds to the variables
        :math:`r` and :math:`\\Delta r` respectively for interactions involving the central atom,
        and ``interaction_cutoff`` and ``interaction_decay`` corresponds to the variables
        :math:`r` and :math:`\\Delta r` respectively for interactions not involving the central atom.

        if ``sorting = 'row-norm'``, the atom indices are ordered such that

            :math:`\\sum_j M_{1j}(k)^2 \\geq \\sum_j M_{2j}(k)^2 \\geq ... \\geq \\sum_j M_{nj}(k)^2`

        if ``sorting = 'distance'``, the atom indices are ordered such that

        .. math::

            \\|{\\bf R}_{1} - {\\bf R}_{k}\\| \\leq \\|{\\bf R}_{2} - {\\bf R}_{k}\\|
                \\leq ... \\leq \\|{\\bf R}_{n} - {\\bf R}_{k}\\|

        The upper triangular of M, including the diagonal, is concatenated to a 1D
        vector representation.

        The representation can be calculated for a subset by either specifying
        ``indices = [0,1,...]``, where :math:`[0,1,...]` are the requested atom indices,
        or by specifying ``indices = 'C'`` to only calculate central carbon atoms.

        The representation is calculated using an OpenMP parallel Fortran routine.

        :param nuclear_charges: Nuclear charges of the atoms in the molecule
        :type nuclear_charges: numpy array
        :param coordinates: 3D Coordinates of the atoms in the molecule
        :type coordinates: numpy array
        :param size: The size of the largest molecule supported by the representation
        :type size: integer
        :param sorting: How the atom indices are sorted ('row-norm', 'distance')
        :type sorting: string
        :param central_cutoff: The distance from the central atom, where the coulomb interaction
            element will be zero
        :type central_cutoff: float
        :param central_decay: The distance over which the the coulomb interaction decays from full to none
        :type central_decay: float
        :param interaction_cutoff: The distance between two non-central atom, where the coulomb interaction
            element will be zero
        :type interaction_cutoff: float
        :param interaction_decay: The distance over which the the coulomb interaction decays from full to none
        :type interaction_decay: float
        :param indices: Subset indices or atomtype
        :type indices: Nonetype/array/string


        :return: nD representation - shape (:math:`N_{atoms}`, size(size+1)/2)
        :rtype: numpy array
    """

    if indices is None:
        nindices = len(nuclear_charges)
        indices = np.arange(1, 1 + nindices, 1, dtype=int)
    elif isinstance(indices, str):
        if indices in NUCLEAR_CHARGE:
            indices = np.where(nuclear_charges == NUCLEAR_CHARGE[indices])[0] + 1
            nindices = indices.size
            if nindices == 0:
                return np.zeros((0, 0))

        else:
            raise ValueError("Unknown value %s given for 'indices' variable" % indices)
    else:
        indices = np.asarray(indices, dtype=int) + 1
        nindices = indices.size

    if sorting == "row-norm":
        return fgenerate_local_coulomb_matrix(
            indices,
            nindices,
            nuclear_charges,
            coordinates,
            nuclear_charges.size,
            size,
            central_cutoff,
            central_decay,
            interaction_cutoff,
            interaction_decay,
        )

    elif sorting == "distance":
        return fgenerate_atomic_coulomb_matrix(
            indices,
            nindices,
            nuclear_charges,
            coordinates,
            nuclear_charges.size,
            size,
            central_cutoff,
            central_decay,
            interaction_cutoff,
            interaction_decay,
        )

    else:
        raise ValueError("Unknown sorting scheme requested")


def generate_coulomb_matrix_eigenvalue(
    nuclear_charges: ndarray, coordinates: ndarray, size: int = 23
) -> ndarray:
    """ Creates an eigenvalue Coulomb Matrix representation of a molecule.
        A matrix :math:`M` is constructed with elements

        .. math::

            M_{ij} =
              \\begin{cases}
                 \\tfrac{1}{2} Z_{i}^{2.4} & \\text{if } i = j \\\\
                 \\frac{Z_{i}Z_{j}}{\\| {\\bf R}_{i} - {\\bf R}_{j}\\|}       & \\text{if } i \\neq j
              \\end{cases},

        where :math:`i` and :math:`j` are atom indices, :math:`Z` is nuclear charge and
        :math:`\\bf R` is the coordinate in euclidean space.
        The molecular representation of the molecule is then the sorted eigenvalues of M.
        The representation is calculated using an OpenMP parallel Fortran routine.

        :param nuclear_charges: Nuclear charges of the atoms in the molecule
        :type nuclear_charges: numpy array
        :param coordinates: 3D Coordinates of the atoms in the molecule
        :type coordinates: numpy array
        :param size: The size of the largest molecule supported by the representation
        :type size: integer

        :return: 1D representation - shape (size, )
        :rtype: numpy array
    """
    return fgenerate_eigenvalue_coulomb_matrix(nuclear_charges, coordinates, size)


def generate_bob(
    nuclear_charges: ndarray,
    coordinates: ndarray,
    atomtypes: ndarray,
    size: int = 23,
    asize: Dict[str, Union[int64, int]] = {"O": 3, "C": 7, "N": 3, "H": 16, "S": 1},
) -> ndarray:
    """Creates a Bag of Bonds (BOB) representation of a molecule.
    The representation expands on the coulomb matrix representation.
    For each element a bag (vector) is constructed for self interactions
    (e.g. ('C', 'H', 'O')).
    For each element pair a bag is constructed for interatomic interactions
    (e.g. ('CC', 'CH', 'CO', 'HH', 'HO', 'OO')), sorted by value.
    The self interaction of element :math:`I` is given by

        :math:`\\tfrac{1}{2} Z_{I}^{2.4}`,

    with :math:`Z_{i}` being the nuclear charge of element :math:`i`
    The interaction between atom :math:`i` of element :math:`I` and
    atom :math:`j` of element :math:`J` is given by

        :math:`\\frac{Z_{I}Z_{J}}{\\| {\\bf R}_{i} - {\\bf R}_{j}\\|}`

    with :math:`R_{i}` being the euclidean coordinate of atom :math:`i`.
    The sorted bags are concatenated to an 1D vector representation.
    The representation is calculated using an OpenMP parallel Fortran routine.

    :param nuclear_charges: Nuclear charges of the atoms in the molecule
    :type nuclear_charges: numpy array
    :param coordinates: 3D Coordinates of the atoms in the molecule
    :type coordinates: numpy array
    :param size: The maximum number of atoms in the representation
    :type size: integer
    :param asize: The maximum number of atoms of each element type supported by the representation
    :type asize: dictionary

    :return: 1D representation
    :rtype: numpy array
    """

    # TODO Moving between str and int is _, should translate everything to use int

    n = 0
    atoms = sorted(asize, key=asize.get)
    nmax = [asize[key] for key in atoms]
    ids = np.zeros(len(nmax), dtype=int)
    for i, (key, value) in enumerate(zip(atoms, nmax)):
        n += value * (1 + value)
        ids[i] = NUCLEAR_CHARGE[key]
        for j in range(i):
            v = nmax[j]
            n += 2 * value * v
    n /= 2

    return fgenerate_bob(nuclear_charges, coordinates, nuclear_charges, ids, nmax, n)


def get_slatm_mbtypes(nuclear_charges: List[ndarray], pbc: str = "000") -> List[List[int64]]:
    """
    Get the list of minimal types of many-body terms in a dataset. This resulting list
    is necessary as input in the ``generate_slatm()`` function.

    :param nuclear_charges: A list of the nuclear charges for each compound in the dataset.
    :type nuclear_charges: list of numpy arrays
    :param pbc: periodic boundary condition along x,y,z direction, defaulted to '000', i.e., molecule
    :type pbc: string
    :return: A list containing the types of many-body terms.
    :rtype: list
    """

    zs = nuclear_charges

    nm = len(zs)
    zsmax = set()
    nas = []
    zs_ravel = []
    for zsi in zs:
        na = len(zsi)
        nas.append(na)
        zsil = list(zsi)
        zs_ravel += zsil
        zsmax.update(zsil)

    zsmax = np.array(list(zsmax))
    nass = []
    for i in range(nm):
        zsi = np.array(zs[i], np.int32)
        nass.append([(zi == zsi).sum() for zi in zsmax])

    nzmax = np.max(np.array(nass), axis=0)
    nzmax_u = []
    if pbc != "000":
        # the PBC will introduce new many-body terms, so set
        # nzmax to 3 if it's less than 3
        for nzi in nzmax:
            if nzi <= 2:
                nzi = 3
            nzmax_u.append(nzi)
        nzmax = nzmax_u

    boas = [
        [
            zi,
        ]
        for zi in zsmax
    ]

    bops = [[zi, zi] for zi in zsmax] + [list(x) for x in itertools.combinations(zsmax, 2)]

    bots = []
    for i in zsmax:
        for bop in bops:
            j, k = bop
            tas = [[i, j, k], [i, k, j], [j, i, k]]
            for tasi in tas:
                if (tasi not in bots) and (tasi[::-1] not in bots):
                    nzsi = [(zj == tasi).sum() for zj in zsmax]
                    if np.all(nzsi <= nzmax):
                        bots.append(tasi)
    mbtypes = boas + bops + bots

    return mbtypes  # , np.array(zs_ravel), np.array(nas)


def generate_slatm(
    nuclear_charges: ndarray,
    coordinates: ndarray,
    mbtypes: List[List[int64]],
    unit_cell: None = None,
    local: bool = False,
    sigmas: List[float] = [0.05, 0.05],
    dgrids: List[float] = [0.03, 0.03],
    rcut: float = 4.8,
    alchemy: bool = False,
    pbc: str = "000",
    rpower: int = 6,
) -> Union[ndarray, List[ndarray]]:
    """
    Generate Spectrum of London and Axillrod-Teller-Muto potential (SLATM) representation.
    Both global (``local=False``) and local (``local=True``) SLATM are available.

    A version that works for periodic boundary conditions will be released soon.

    NOTE: You will need to run the ``get_slatm_mbtypes()`` function to get the ``mbtypes`` input (or generate it manually).

    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param coordinates: Input coordinates
    :type coordinates: numpy array
    :param mbtypes: Many-body types for the whole dataset, including 1-, 2- and 3-body types. Could be obtained by calling ``get_slatm_mbtypes()``.
    :type mbtypes: list
    :param local: Generate a local representation. Defaulted to False (i.e., global representation); otherwise, atomic version.
    :type local: bool
    :param sigmas: Controlling the width of Gaussian smearing function for 2- and 3-body parts, defaulted to [0.05,0.05], usually these do not need to be adjusted.
    :type sigmas: list
    :param dgrids: The interval between two sampled internuclear distances and angles, defaulted to [0.03,0.03], no need for change, compromised for speed and accuracy.
    :type dgrids: list
    :param rcut: Cut-off radius, defaulted to 4.8 Angstrom.
    :type rcut: float
    :param alchemy: Swith to use the alchemy version of SLATM. (default=False)
    :type alchemy: bool
    :param pbc: defaulted to '000', meaning it's a molecule; the three digits in the string corresponds to x,y,z direction
    :type pbc: string
    :param rpower: The power of R in 2-body potential, defaulted to London potential (=6).
    :type rpower: float
    :return: 1D SLATM representation
    :rtype: numpy array
    """

    c = unit_cell
    # UNUSED iprt = False
    if c is None:
        c = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    if pbc != "000":
        if c is None:
            raise ValueError("Please specify unit cell for SLATM")

        # =======================================================================
        # PBC may introduce new many-body terms, so at the stage of get statistics
        # info from db, we've already considered this point by letting maximal number
        # of nuclear charges being 3.
        # =======================================================================

    zs = nuclear_charges
    na = len(zs)
    coords = coordinates
    obj = [zs, coords, c]

    iloc = local

    if iloc:
        mbs = []
        X2Ns = []
        for ia in range(na):
            n1 = 0
            n2 = 0
            n3 = 0
            mbs_ia = np.zeros(0)
            # UNUSED icount = 0
            for mbtype in mbtypes:
                if len(mbtype) == 1:
                    mbsi = get_boa(
                        mbtype[0],
                        np.array(
                            [
                                zs[ia],
                            ]
                        ),
                    )
                    if alchemy:
                        n1 = 1
                        n1_0 = mbs_ia.shape[0]
                        if n1_0 == 0:
                            mbs_ia = np.concatenate((mbs_ia, mbsi), axis=0)
                        elif n1_0 == 1:
                            mbs_ia += mbsi
                        else:
                            raise ValueError()
                    else:
                        n1 += len(mbsi)
                        mbs_ia = np.concatenate((mbs_ia, mbsi), axis=0)
                elif len(mbtype) == 2:
                    mbsi = get_sbop(
                        mbtype,
                        obj,
                        iloc=iloc,
                        ia=ia,
                        sigma=sigmas[0],
                        dgrid=dgrids[0],
                        rcut=rcut,
                        pbc=pbc,
                        rpower=rpower,
                    )
                    mbsi *= 0.5  # only for the two-body parts, local rpst
                    if alchemy:
                        n2 = len(mbsi)
                        n2_0 = mbs_ia.shape[0]
                        if n2_0 == n1:
                            mbs_ia = np.concatenate((mbs_ia, mbsi), axis=0)
                        elif n2_0 == n1 + n2:
                            t = mbs_ia[n1 : n1 + n2] + mbsi
                            mbs_ia[n1 : n1 + n2] = t
                        else:
                            raise ValueError()
                    else:
                        n2 += len(mbsi)
                        mbs_ia = np.concatenate((mbs_ia, mbsi), axis=0)
                else:  # len(mbtype) == 3:
                    mbsi = get_sbot(
                        mbtype,
                        obj,
                        iloc=iloc,
                        ia=ia,
                        sigma=sigmas[1],
                        dgrid=dgrids[1],
                        rcut=rcut,
                        pbc=pbc,
                    )

                    if alchemy:
                        n3 = len(mbsi)
                        n3_0 = mbs_ia.shape[0]
                        if n3_0 == n1 + n2:
                            mbs_ia = np.concatenate((mbs_ia, mbsi), axis=0)
                        elif n3_0 == n1 + n2 + n3:
                            t = mbs_ia[n1 + n2 : n1 + n2 + n3] + mbsi
                            mbs_ia[n1 + n2 : n1 + n2 + n3] = t
                        else:
                            raise ValueError()
                    else:
                        n3 += len(mbsi)
                        mbs_ia = np.concatenate((mbs_ia, mbsi), axis=0)

            mbs.append(mbs_ia)
            X2N = [n1, n2, n3]
            if X2N not in X2Ns:
                X2Ns.append(X2N)

        if len(X2Ns) != 1:
            raise ValueError("multiple `X2N ???")

    else:
        n1 = 0
        n2 = 0
        n3 = 0
        mbs = np.zeros(0)
        for mbtype in mbtypes:
            if len(mbtype) == 1:
                mbsi = get_boa(mbtype[0], zs)
                if alchemy:
                    n1 = 1
                    n1_0 = mbs.shape[0]
                    if n1_0 == 0:
                        mbs = np.concatenate((mbs, [sum(mbsi)]), axis=0)
                    elif n1_0 == 1:
                        mbs += sum(mbsi)
                    else:
                        raise ValueError()
                else:
                    n1 += len(mbsi)
                    mbs = np.concatenate((mbs, mbsi), axis=0)
            elif len(mbtype) == 2:
                mbsi = get_sbop(
                    mbtype, obj, sigma=sigmas[0], dgrid=dgrids[0], rcut=rcut, rpower=rpower
                )

                if alchemy:
                    n2 = len(mbsi)
                    n2_0 = mbs.shape[0]
                    if n2_0 == n1:
                        mbs = np.concatenate((mbs, mbsi), axis=0)
                    elif n2_0 == n1 + n2:
                        t = mbs[n1 : n1 + n2] + mbsi
                        mbs[n1 : n1 + n2] = t
                    else:
                        raise ValueError()
                else:
                    n2 += len(mbsi)
                    mbs = np.concatenate((mbs, mbsi), axis=0)
            else:  # len(mbtype) == 3:
                mbsi = get_sbot(mbtype, obj, sigma=sigmas[1], dgrid=dgrids[1], rcut=rcut)

                if alchemy:
                    n3 = len(mbsi)
                    n3_0 = mbs.shape[0]
                    if n3_0 == n1 + n2:
                        mbs = np.concatenate((mbs, mbsi), axis=0)
                    elif n3_0 == n1 + n2 + n3:
                        t = mbs[n1 + n2 : n1 + n2 + n3] + mbsi
                        mbs[n1 + n2 : n1 + n2 + n3] = t
                    else:
                        raise ValueError()
                else:
                    n3 += len(mbsi)
                    mbs = np.concatenate((mbs, mbsi), axis=0)

    return mbs


def generate_acsf(
    nuclear_charges: List[int],
    coordinates: ndarray,
    elements: List[int] = [1, 6, 7, 8, 16],
    nRs2: int = 3,
    nRs3: int = 3,
    nTs: int = 3,
    eta2: int = 1,
    eta3: int = 1,
    zeta: int = 1,
    rcut: int = 5,
    acut: int = 5,
    bin_min: float = 0.8,
    gradients: bool = False,
    pad: Optional[int] = None,
) -> Union[Tuple[ndarray, ndarray], ndarray]:
    """
    Generate the variant of atom-centered symmetry functions used in https://doi.org/10.1039/C7SC04934J

    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param coordinates: Input coordinates
    :type coordinates: numpy array
    :param elements: list of unique nuclear charges (atom types)
    :type elements: numpy array
    :param nRs2: Number of gaussian basis functions in the two-body terms
    :type nRs2: integer
    :param nRs3: Number of gaussian basis functions in the three-body radial part
    :type nRs3: integer
    :param nTs: Number of basis functions in the three-body angular part
    :type nTs: integer
    :param eta2: Precision in the gaussian basis functions in the two-body terms
    :type eta2: float
    :param eta3: Precision in the gaussian basis functions in the three-body radial part
    :type eta3: float
    :param zeta: Precision parameter of basis functions in the three-body angular part
    :type zeta: float
    :param rcut: Cut-off radius of the two-body terms
    :type rcut: float
    :param acut: Cut-off radius of the three-body terms
    :type acut: float
    :param bin_min: the value at which to start binning the distances
    :type bin_min: positive float
    :param gradients: To return gradients or not
    :type gradients: boolean
    :param pad: `None` if no padding is to be applied other, otherwise an integer corresponding to the desired size
    :type gradients: NoneType or integer
    :return: Atom-centered symmetry functions representation
    :rtype: numpy array
    """

    Rs2 = np.linspace(bin_min, rcut, nRs2)
    Rs3 = np.linspace(bin_min, acut, nRs3)
    Ts = np.linspace(0, np.pi, nTs)
    n_elements = len(elements)
    natoms = len(coordinates)

    descr_size = n_elements * nRs2 + (n_elements * (n_elements + 1)) // 2 * nRs3 * nTs

    if gradients is False:

        rep = fgenerate_acsf(
            coordinates,
            nuclear_charges,
            elements,
            Rs2,
            Rs3,
            Ts,
            eta2,
            eta3,
            zeta,
            rcut,
            acut,
            natoms,
            descr_size,
        )

        if pad is not None:

            rep_pad = np.zeros((pad, descr_size))
            rep_pad[:natoms, :] += rep

            return rep_pad

        else:
            return rep

    else:

        (rep, grad) = fgenerate_acsf_and_gradients(
            coordinates,
            nuclear_charges,
            elements,
            Rs2,
            Rs3,
            Ts,
            eta2,
            eta3,
            zeta,
            rcut,
            acut,
            natoms,
            descr_size,
        )

        if pad is not None:
            rep_pad = np.zeros((pad, descr_size))
            grad_pad = np.zeros((pad, descr_size, pad, 3))

            rep_pad[:natoms, :] += rep
            grad_pad[:natoms, :, :natoms, :] += grad

            return rep_pad, grad_pad
        else:
            return rep, grad


def generate_fchl19(
    nuclear_charges: ndarray,
    coordinates: ndarray,
    elements: List[int] = [1, 6, 7, 8, 16],
    nRs2: int = 24,
    nRs3: int = 20,
    nFourier: int = 1,
    eta2: float = 0.32,
    eta3: float = 2.7,
    zeta: float = np.pi,
    rcut: float = 8.0,
    acut: float = 8.0,
    two_body_decay: float = 1.8,
    three_body_decay: float = 0.57,
    three_body_weight: float = 13.4,
    pad: Union[int, bool] = False,
    gradients: bool = False,
) -> Union[Tuple[ndarray, ndarray], ndarray]:
    """
    FCHL-ACSF

    https://pubs.aip.org/aip/jcp/article/152/4/044107/1064737/FCHL-revisited-Faster-and-more-accurate-quantum

    Reasonable hyperparameters:

    Sigma ~ 21.0
    Lambda ~ 1e-8
    Max singular value ~ 1e-12

    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param coordinates: Input coordinates
    :type coordinates: numpy array
    :param elements: list of unique nuclear charges (atom types)
    :type elements: numpy array
    :param nRs2: Number of gaussian basis functions in the two-body terms
    :type nRs2: integer
    :param nRs3: Number of gaussian basis functions in the three-body radial part
    :type nRs3: integer
    :param nFourier: Order of Fourier expansion
    :type nFourier: integer
    :param eta2: Precision in the gaussian basis functions in the two-body terms
    :type eta2: float
    :param eta3: Precision in the gaussian basis functions in the three-body radial part
    :type eta3: float
    :param zeta: Precision parameter of basis functions in the three-body angular part
    :type zeta: float
    :param rcut: Cut-off radius of the two-body terms
    :type rcut: float
    :param acut: Cut-off radius of the three-body terms
    :type acut: float
    :param gradients: To return gradients or not
    :type gradients: boolean
    :return: Atom-centered symmetry functions representation
    :rtype: numpy array
    """

    Rs2 = np.linspace(0, rcut, 1 + nRs2)[1:]
    Rs3 = np.linspace(0, acut, 1 + nRs3)[1:]

    Ts = np.linspace(0, np.pi, 2 * nFourier)
    n_elements = len(elements)
    natoms = len(coordinates)

    descr_size = n_elements * nRs2 + (n_elements * (n_elements + 1)) * nRs3 * nFourier

    # Normalization constant for three-body
    three_body_weight = np.sqrt(eta3 / np.pi) * three_body_weight

    if gradients is False:

        rep = fgenerate_fchl_acsf(
            coordinates,
            nuclear_charges,
            elements,
            Rs2,
            Rs3,
            Ts,
            eta2,
            eta3,
            zeta,
            rcut,
            acut,
            natoms,
            descr_size,
            two_body_decay,
            three_body_decay,
            three_body_weight,
        )

        if pad is not False:

            rep_pad = np.zeros((pad, descr_size))
            rep_pad[:natoms, :] += rep

            return rep_pad

        else:
            return rep

    else:

        if nFourier > 1:
            raise ValueError(f"FCHL-ACSF only supports nFourier=1, requested {nFourier}")

        (rep, grad) = fgenerate_fchl_acsf_and_gradients(
            coordinates,
            nuclear_charges,
            elements,
            Rs2,
            Rs3,
            Ts,
            eta2,
            eta3,
            zeta,
            rcut,
            acut,
            natoms,
            descr_size,
            two_body_decay,
            three_body_decay,
            three_body_weight,
        )

        if pad is not False:
            rep_pad = np.zeros((pad, descr_size))
            grad_pad = np.zeros((pad, descr_size, pad, 3))

            rep_pad[:natoms, :] += rep
            grad_pad[:natoms, :, :natoms, :] += grad

            return rep_pad, grad_pad
        else:
            return rep, grad



import numpy as np
from conftest import ASSETS, get_energies, shuffle_arrays

from qmllib.kernels import get_local_kernels_gaussian, get_local_kernels_laplacian
from qmllib.representations import generate_coulomb_matrix_atomic
from qmllib.solvers import cho_solve
from qmllib.utils.xyz_format import read_xyz


def test_krr_gaussian_local_cmat():

    # Parse file containing PBE0/def2-TZVP heats of formation and xyz filenames
    data = get_energies(ASSETS / "hof_qm7.txt")

    n_points = 1000

    all_representations = []
    all_properties = []
    all_atoms = []

    filenames = sorted(data.keys())[:n_points]

    for filename in filenames:
        coord, atoms = read_xyz((ASSETS / "qm7" / filename).with_suffix(".xyz"))

        representation = generate_coulomb_matrix_atomic(atoms, coord, size=23, sorting="row-norm")

        all_representations.append(representation)
        all_properties.append(data[filename])
        all_atoms.append(atoms)

    all_properties = np.array(all_properties)

    shuffle_arrays(all_atoms, all_properties, all_representations, seed=666)

    # Make training and test sets
    n_test = 100
    n_train = 200
    indices = list(range(n_points))
    train_indices = indices[:n_train]
    test_indices = indices[-n_test:]

    train_representations = np.concatenate([all_representations[i] for i in train_indices])
    test_representations = np.concatenate([all_representations[i] for i in test_indices])

    test_atoms = [all_atoms[x] for x in test_indices]
    test_properties = all_properties[test_indices]

    train_atoms = [all_atoms[x] for x in train_indices]
    train_properties = all_properties[train_indices]

    train_sizes = np.array([len(atoms) for atoms in train_atoms])
    test_sizes = np.array([len(atoms) for atoms in test_atoms])

    # Set hyper-parameters
    sigma = 724.0
    llambda = 10 ** (-6.5)

    K = get_local_kernels_gaussian(
        train_representations, train_representations, train_sizes, train_sizes, [sigma]
    )[0]
    assert np.allclose(K, K.T), "Error in local Gaussian kernel symmetry"

    K_test = np.loadtxt(ASSETS / "K_local_gaussian.txt")
    assert np.allclose(K, K_test), "Error in local Gaussian kernel (vs. reference)"

    # Solve alpha
    K[np.diag_indices_from(K)] += llambda
    alpha = cho_solve(K, train_properties)

    print(test_representations.shape)
    print(test_sizes)

    # Calculate prediction kernel
    Ks = get_local_kernels_gaussian(
        test_representations, train_representations, test_sizes, train_sizes, [sigma]
    )[0]

    Ks_test = np.loadtxt(ASSETS / "Ks_local_gaussian.txt")
    # Somtimes a few coulomb matrices differ because of parallel sorting and numerical error
    # Allow up to 5 molecules to differ from the supplied reference.
    differences_count = len(set(np.where(Ks - Ks_test > 1e-7)[0]))
    assert differences_count < 5, "Error in local Laplacian kernel (vs. reference)"
    # assert np.allclose(Ks, Ks_test), "Error in local Gaussian kernel (vs. reference)"

    predicted_properties = np.dot(Ks, alpha)

    mae = np.mean(np.abs(test_properties - predicted_properties))
    print(mae)
    assert abs(19.0 - mae) < 1.0, "Error in local Gaussian kernel-ridge regression"


def test_krr_laplacian_local_cmat():

    # Parse file containing PBE0/def2-TZVP heats of formation and xyz filenames
    data = get_energies(ASSETS / "hof_qm7.txt")

    n_points = 1000

    all_representations = []
    all_properties = []
    all_atoms = []

    filenames = sorted(data.keys())[:n_points]

    for filename in filenames:
        coord, atoms = read_xyz((ASSETS / "qm7" / filename).with_suffix(".xyz"))

        representation = generate_coulomb_matrix_atomic(atoms, coord, size=23, sorting="row-norm")

        all_representations.append(representation)
        all_properties.append(data[filename])
        all_atoms.append(atoms)

    all_properties = np.array(all_properties)

    shuffle_arrays(all_atoms, all_properties, all_representations, seed=666)

    # Make training and test sets
    n_test = 100
    n_train = 200
    indices = list(range(n_points))
    train_indices = indices[:n_train]
    test_indices = indices[-n_test:]

    train_representations = np.concatenate([all_representations[i] for i in train_indices])
    test_representations = np.concatenate([all_representations[i] for i in test_indices])

    test_atoms = [all_atoms[x] for x in test_indices]
    test_properties = all_properties[test_indices]

    train_atoms = [all_atoms[x] for x in train_indices]
    train_properties = all_properties[train_indices]

    train_sizes = np.array([len(atoms) for atoms in train_atoms])
    test_sizes = np.array([len(atoms) for atoms in test_atoms])

    # Set hyper-parameters
    sigma = 10 ** (3.6)
    llambda = 10 ** (-12.0)

    K = get_local_kernels_laplacian(
        train_representations, train_representations, train_sizes, train_sizes, [sigma]
    )[0]
    assert np.allclose(K, K.T), "Error in local Laplacian kernel symmetry"

    K_test = np.loadtxt(ASSETS / "K_local_laplacian.txt")
    assert np.allclose(K, K_test), "Error in local Laplacian kernel (vs. reference)"

    # Solve alpha
    K[np.diag_indices_from(K)] += llambda
    alpha = cho_solve(K, train_properties)

    # Calculate prediction kernel
    Ks = get_local_kernels_laplacian(
        test_representations, train_representations, test_sizes, train_sizes, [sigma]
    )[0]

    Ks_test = np.loadtxt(ASSETS / "Ks_local_laplacian.txt")

    # Somtimes a few coulomb matrices differ because of parallel sorting and numerical error
    # Allow up to 5 molecules to differ from the supplied reference.
    differences_count = len(set(np.where(Ks - Ks_test > 1e-7)[0]))
    assert differences_count < 5, "Error in local Laplacian kernel (vs. reference)"

    predicted_properties = np.dot(Ks, alpha)

    mae = np.mean(np.abs(test_properties - predicted_properties))
    assert abs(8.7 - mae) < 1.0, "Error in local Laplacian kernel-ridge regression"



from pathlib import Path
from typing import Tuple

import numpy as np
from numpy import ndarray

from qmllib.constants.periodic_table import NUCLEAR_CHARGE


def read_xyz(filename: str | Path) -> Tuple[ndarray, ndarray]:
    """(Re-)initializes the Compound-object with data from an xyz-file.

    :param filename: Input xyz-filename or file-like obejct
    :type filename: string or file-like object
    """

    # if isinstance(filename, string_types):
    #     with open(filename, "r") as f:
    #         lines = f.readlines()
    # else:
    #     lines = filename.readlines()

    with open(filename, "r") as f:
        lines = f.readlines()

    natoms = int(lines[0])
    atomtypes = []
    nuclear_charges = np.empty(natoms, dtype=int)
    coordinates = np.empty((natoms, 3), dtype=float)

    # Give the Compound a name if it is a string
    # name = filename if isinstance(filename, string_types) else "Compound"

    for i, line in enumerate(lines[2 : natoms + 2]):
        tokens = line.split()

        if len(tokens) < 4:
            break

        atomtypes.append(tokens[0])
        nuclear_charges[i] = NUCLEAR_CHARGE[tokens[0]]

        coordinates[i] = np.asarray(tokens[1:4], dtype=float)

    return coordinates, nuclear_charges


# self.natypes = dict([(key, len(value)) for key,value in self.atomtype_indices.items()])



from typing import List

import numpy as np
from numpy import ndarray

from qmllib.utils.alchemy import PTP

from .farad_kernels import (
    fget_atomic_kernels_arad,
    fget_atomic_symmetric_kernels_arad,
    fget_global_kernels_arad,
    fget_global_symmetric_kernels_arad,
    fget_local_kernels_arad,
    fget_local_symmetric_kernels_arad,
)


def getAngle(sp: ndarray, norms: ndarray) -> ndarray:
    epsilon = 10.0 * np.finfo(float).eps
    angles = np.zeros(sp.shape)
    mask1 = np.logical_and(np.abs(sp - norms) > epsilon, np.abs(norms) > epsilon)
    angles[mask1] = np.arccos(np.clip(sp[mask1] / norms[mask1], -1.0, 1.0))
    return angles


def generate_arad(
    nuclear_charges: ndarray, coordinates: ndarray, size: int = 23, cut_distance: float = 5.0
) -> ndarray:
    """Generates a representation for the ARAD kernel module.

    :param coordinates: Input coordinates.
    :type coordinates: numpy array
    :param nuclear_charges: List of nuclear charges.
    :type nuclear_charges: numpy array
    :param size: Max number of atoms in representation.
    :type size: integer
    :param cut_distance: Spatial cut-off distance.
    :type cut_distance: float
    :return: ARAD representation, shape = (size,5,size).
    :rtype: numpy array
    """

    # PBC is not supported by the kernels currently
    cell = None
    maxAts = size
    maxMolSize = size

    coords = coordinates
    occupationList = nuclear_charges
    cut = cut_distance

    L = coords.shape[0]
    occupationList = np.asarray(occupationList)
    M = np.zeros((maxMolSize, 5, maxAts))

    if cell is not None:
        coords = np.dot(coords, cell)
        nExtend = (np.floor(cut / np.linalg.norm(cell, 2, axis=0)) + 1).astype(int)
        for i in range(-nExtend[0], nExtend[0] + 1):
            for j in range(-nExtend[1], nExtend[1] + 1):
                for k in range(-nExtend[2], nExtend[2] + 1):
                    if i == -nExtend[0] and j == -nExtend[1] and k == -nExtend[2]:
                        coordsExt = coords + i * cell[0, :] + j * cell[1, :] + k * cell[2, :]
                        occupationListExt = occupationList.copy()
                    else:
                        occupationListExt = np.append(occupationListExt, occupationList)
                        coordsExt = np.append(
                            coordsExt,
                            coords + i * cell[0, :] + j * cell[1, :] + k * cell[2, :],
                            axis=0,
                        )

    else:
        coordsExt = coords.copy()
        occupationListExt = occupationList.copy()

    M[:, 0, :] = 1e100

    for i in range(L):
        # Calculate Distance
        cD = coordsExt[:] - coords[i]
        ocExt = np.asarray([PTP[o] for o in occupationListExt])

        # Obtaining angles
        sp = np.sum(cD[:, np.newaxis] * cD[np.newaxis, :], axis=2)
        D1 = np.sqrt(np.sum(cD**2, axis=1))
        D2 = D1[:, np.newaxis] * D1[np.newaxis, :]
        angs = getAngle(sp, D2)

        # Obtaining cos and sine terms
        cosAngs = np.cos(angs) * (1.0 - np.sin(np.pi * D1[np.newaxis, :] / (2.0 * cut)))
        sinAngs = np.sin(angs) * (1.0 - np.sin(np.pi * D1[np.newaxis, :] / (2.0 * cut)))

        args = np.argsort(D1)

        D1 = D1[args]

        ocExt = np.asarray([ocExt[l] for l in args])

        sub_indices = np.ix_(args, args)
        cosAngs = cosAngs[sub_indices]
        sinAngs = sinAngs[sub_indices]

        args = np.where(D1 < cut)[0]

        D1 = D1[args]

        ocExt = np.asarray([ocExt[l] for l in args])

        sub_indices = np.ix_(args, args)
        cosAngs = cosAngs[sub_indices]
        sinAngs = sinAngs[sub_indices]

        norm = np.sum(1.0 - np.sin(np.pi * D1[np.newaxis, :] / (2.0 * cut)))
        M[i, 0, : len(D1)] = D1
        M[i, 1, : len(D1)] = ocExt[:, 0]
        M[i, 2, : len(D1)] = ocExt[:, 1]
        M[i, 3, : len(D1)] = np.sum(cosAngs, axis=1) / norm
        M[i, 4, : len(D1)] = np.sum(sinAngs, axis=1) / norm

    return M


def get_global_kernels_arad(
    X1: ndarray,
    X2: ndarray,
    sigmas: List[float],
    width: float = 0.2,
    cut_distance: float = 5.0,
    r_width: float = 1.0,
    c_width: float = 0.5,
) -> ndarray:
    """Calculates the global Gaussian kernel matrix K for atomic ARAD
    descriptors for a list of different sigmas. Each kernel element
    is the sum of all kernel elements between pairs of atoms in two molecules.

    K is calculated using an OpenMP parallel Fortran routine.

    :param X1: ARAD descriptors for molecules in set 1.
    :type X1: numpy array
    :param X2: Array of ARAD descriptors for molecules in set 2.
    :type X2: numpy array
    :param sigmas: List of sigmas for which to calculate the Kernel matrices.
    :type sigmas: list

    :return: The kernel matrices for each sigma - shape (number_sigmas, number_molecules1, number_molecules2)
    :rtype: numpy array
    """

    amax = X1.shape[1]

    if not X1.shape[3] == amax:
        raise ValueError("Check ARAD decriptor sizes")
    if not X2.shape[1] == amax:
        raise ValueError("Check ARAD decriptor sizes")
    if not X2.shape[3] == amax:
        raise ValueError("Check ARAD decriptor sizes")

    nm1 = X1.shape[0]
    nm2 = X2.shape[0]

    N1 = np.empty(nm1, dtype=np.int32)
    Z1_arad = np.zeros((nm1, amax, 2))
    for i in range(nm1):
        N1[i] = len(np.where(X1[i, :, 2, 0] > 0)[0])
        Z1_arad[i] = X1[i, :, 1:3, 0]

    N2 = np.empty(nm2, dtype=np.int32)
    Z2_arad = np.zeros((nm2, amax, 2))
    for i in range(nm2):
        N2[i] = len(np.where(X2[i, :, 2, 0] > 0)[0])
        Z2_arad[i] = X2[i, :, 1:3, 0]

    sigmas = np.array(sigmas)
    nsigmas = sigmas.size

    return fget_global_kernels_arad(
        X1,
        X2,
        Z1_arad,
        Z2_arad,
        N1,
        N2,
        sigmas,
        nm1,
        nm2,
        nsigmas,
        width,
        cut_distance,
        r_width,
        c_width,
    )


def get_global_symmetric_kernels_arad(
    X1: ndarray,
    sigmas: List[float],
    width: float = 0.2,
    cut_distance: float = 5.0,
    r_width: float = 1.0,
    c_width: float = 0.5,
) -> ndarray:
    """Calculates the global Gaussian kernel matrix K for atomic ARAD
    descriptors for a list of different sigmas. Each kernel element
    is the sum of all kernel elements between pairs of atoms in two molecules.

    K is calculated using an OpenMP parallel Fortran routine.

    :param X1: ARAD descriptors for molecules in set 1.
    :type X1: numpy array
    :param sigmas: List of sigmas for which to calculate the Kernel matrices.
    :type sigmas: list

    :return: The kernel matrices for each sigma - shape (number_sigmas, number_molecules1, number_molecules1)
    :rtype: numpy array
    """

    nm1 = X1.shape[0]
    amax = X1.shape[1]

    N1 = np.empty(nm1, dtype=np.int32)
    Z1_arad = np.zeros((nm1, amax, 2))
    for i in range(nm1):
        N1[i] = len(np.where(X1[i, :, 2, 0] > 0)[0])
        Z1_arad[i] = X1[i, :, 1:3, 0]

    sigmas = np.array(sigmas)
    nsigmas = sigmas.size

    return fget_global_symmetric_kernels_arad(
        X1, Z1_arad, N1, sigmas, nm1, nsigmas, width, cut_distance, r_width, c_width
    )


def get_local_kernels_arad(
    X1: ndarray,
    X2: ndarray,
    sigmas: List[float],
    width: float = 0.2,
    cut_distance: float = 5.0,
    r_width: float = 1.0,
    c_width: float = 0.5,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K for atomic ARAD
    descriptors for a list of different sigmas. Each kernel element
    is the sum of all kernel elements between pairs of atoms in two molecules.

    K is calculated using an OpenMP parallel Fortran routine.

    :param X1: ARAD descriptors for molecules in set 1.
    :type X1: numpy array
    :param X2: Array of ARAD descriptors for molecules in set 2.
    :type X2: numpy array
    :param sigmas: List of sigmas for which to calculate the Kernel matrices.
    :type sigmas: list

    :return: The kernel matrices for each sigma - shape (number_sigmas, number_molecules1, number_molecules2)
    :rtype: numpy array
    """

    amax = X1.shape[1]

    if not X1.shape[3] == amax:
        raise ValueError("Check ARAD decriptor sizes")
    if not X2.shape[1] == amax:
        raise ValueError("Check ARAD decriptor sizes")
    if not X2.shape[3] == amax:
        raise ValueError("Check ARAD decriptor sizes")

    nm1 = X1.shape[0]
    nm2 = X2.shape[0]

    N1 = np.empty(nm1, dtype=np.int32)
    Z1_arad = np.zeros((nm1, amax, 2))
    for i in range(nm1):
        N1[i] = len(np.where(X1[i, :, 2, 0] > 0)[0])
        Z1_arad[i] = X1[i, :, 1:3, 0]

    N2 = np.empty(nm2, dtype=np.int32)
    Z2_arad = np.zeros((nm2, amax, 2))
    for i in range(nm2):
        N2[i] = len(np.where(X2[i, :, 2, 0] > 0)[0])
        Z2_arad[i] = X2[i, :, 1:3, 0]

    sigmas = np.array(sigmas)
    nsigmas = sigmas.size

    return fget_local_kernels_arad(
        X1,
        X2,
        Z1_arad,
        Z2_arad,
        N1,
        N2,
        sigmas,
        nm1,
        nm2,
        nsigmas,
        width,
        cut_distance,
        r_width,
        c_width,
    )


def get_local_symmetric_kernels_arad(
    X1: ndarray,
    sigmas: List[float],
    width: float = 0.2,
    cut_distance: float = 5.0,
    r_width: float = 1.0,
    c_width: float = 0.5,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K for atomic ARAD
    descriptors for a list of different sigmas. Each kernel element
    is the sum of all kernel elements between pairs of atoms in two molecules.

    K is calculated using an OpenMP parallel Fortran routine.

    :param X1: ARAD descriptors for molecules in set 1.
    :type X1: numpy array
    :param sigmas: List of sigmas for which to calculate the Kernel matrices.
    :type sigmas: list

    :return: The kernel matrices for each sigma - shape (number_sigmas, number_molecules1, number_molecules1)
    :rtype: numpy array
    """

    nm1 = X1.shape[0]
    amax = X1.shape[1]

    N1 = np.empty(nm1, dtype=np.int32)
    Z1_arad = np.zeros((nm1, amax, 2))
    for i in range(nm1):
        N1[i] = len(np.where(X1[i, :, 2, 0] > 0)[0])
        Z1_arad[i] = X1[i, :, 1:3, 0]

    sigmas = np.array(sigmas)
    nsigmas = sigmas.size

    return fget_local_symmetric_kernels_arad(
        X1, Z1_arad, N1, sigmas, nm1, nsigmas, width, cut_distance, r_width, c_width
    )


def get_atomic_kernels_arad(
    X1: ndarray,
    X2: ndarray,
    sigmas: List[float],
    width: float = 0.2,
    cut_distance: float = 5.0,
    r_width: float = 1.0,
    c_width: float = 0.5,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K for atomic ARAD
    descriptors for a list of different sigmas. For atomic properties, e.g.
    partial charges, chemical shifts, etc.

    K is calculated using an OpenMP parallel Fortran routine.

    :param X1: ARAD descriptors for molecules in set 1. shape=(number_atoms,5,size)
    :type X1: numpy array
    :param X2: ARAD descriptors for molecules in set 1. shape=(number_atoms,5,size)
    :type X2: numpy array
    :param sigmas: List of sigmas for which to calculate the Kernel matrices.
    :type sigmas: list

    :return: The kernel matrices for each sigma - shape (number_sigmas, number_atoms1, number_atoms2)
    :rtype: numpy array
    """

    if not len(X1.shape) == 3:
        raise ValueError("Expected different shape")
    if not len(X2.shape) == 3:
        raise ValueError("Expected different shape")

    na1 = X1.shape[0]
    na2 = X2.shape[0]

    N1 = np.empty(na1, dtype=np.int32)
    N2 = np.empty(na2, dtype=np.int32)

    Z1_arad = np.zeros((na1, 2))
    Z2_arad = np.zeros((na2, 2))

    for i in range(na1):
        N1[i] = len(np.where(X1[i, 0, :] < cut_distance)[0])
        Z1_arad[i] = X1[i, 1:3, 0]

    for i in range(na2):
        N2[i] = len(np.where(X2[i, 0, :] < cut_distance)[0])
        Z2_arad[i] = X2[i, 1:3, 0]

    sigmas = np.array(sigmas)
    nsigmas = sigmas.size

    return fget_atomic_kernels_arad(
        X1,
        X2,
        Z1_arad,
        Z2_arad,
        N1,
        N2,
        sigmas,
        na1,
        na2,
        nsigmas,
        width,
        cut_distance,
        r_width,
        c_width,
    )


def get_atomic_symmetric_kernels_arad(
    X1: ndarray,
    sigmas: List[float],
    width: float = 0.2,
    cut_distance: float = 5.0,
    r_width: float = 1.0,
    c_width: float = 0.5,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K for atomic ARAD
    descriptors for a list of different sigmas. For atomic properties, e.g.
    partial charges, chemical shifts, etc.

    K is calculated using an OpenMP parallel Fortran routine.

    :param X1: ARAD descriptors for molecules in set 1. shape=(number_atoms,5,size)
    :type X1: numpy array
    :param sigmas: List of sigmas for which to calculate the Kernel matrices.
    :type sigmas: list

    :return: The kernel matrices for each sigma - shape (number_sigmas, number_atoms1, number_atoms1)
    :rtype: numpy array
    """

    if not len(X1.shape) == 3:
        raise ValueError("Expected different shape")
    na1 = X1.shape[0]

    N1 = np.empty(na1, dtype=np.int32)
    Z1_arad = np.zeros((na1, 2))

    for i in range(na1):
        N1[i] = len(np.where(X1[i, 0, :] < cut_distance)[0])
        Z1_arad[i] = X1[i, 1:3, 0]

    sigmas = np.array(sigmas)
    nsigmas = sigmas.size

    return fget_atomic_symmetric_kernels_arad(
        X1, Z1_arad, N1, sigmas, na1, nsigmas, width, cut_distance, r_width, c_width
    )



from typing import Union

import numpy as np
from numpy import ndarray

from .fdistance import fl2_distance, fmanhattan_distance, fp_distance_double, fp_distance_integer


def manhattan_distance(A: ndarray, B: ndarray) -> ndarray:
    """Calculates the Manhattan distances, D,  between two
    Numpy arrays of representations.

        :math:`D_{ij} = \\|A_i - B_j\\|_1`

    Where :math:`A_{i}` and :math:`B_{j}` are representation vectors.
    D is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of descriptors - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of descriptors - shape (M, representation size).
    :type B: numpy array

    :return: The Manhattan-distance matrix.
    :rtype: numpy array
    """

    if len(A.shape) != 2 or len(B.shape) != 2:
        raise ValueError("expected matrices of dimension=2")

    if B.shape[1] != A.shape[1]:
        raise ValueError("expected matrices containing vectors of same size")

    na = A.shape[0]
    nb = B.shape[0]

    D = np.empty((na, nb), order="F")

    fmanhattan_distance(A.T, B.T, D)

    return D


def l2_distance(A: ndarray, B: ndarray) -> ndarray:
    """Calculates the L2 distances, D, between two
    Numpy arrays of representations.

        :math:`D_{ij} = \\|A_i - B_j\\|_2`

    Where :math:`A_{i}` and :math:`B_{j}` are representation vectors.
    D is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of descriptors - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of descriptors - shape (M, representation size).
    :type B: numpy array

    :return: The L2-distance matrix.
    :rtype: numpy array
    """

    if len(A.shape) != 2 or len(B.shape) != 2:
        raise ValueError("expected matrices of dimension=2")

    if B.shape[1] != A.shape[1]:
        raise ValueError("expected matrices containing vectors of same size")

    na = A.shape[0]
    nb = B.shape[0]

    D = np.empty((na, nb), order="F")

    fl2_distance(A.T, B.T, D)

    return D


def p_distance(A: ndarray, B: ndarray, p: Union[int, float] = 2) -> ndarray:
    """Calculates the p-norm distances between two
    Numpy arrays of representations.
    The value of the keyword argument ``p =`` sets the norm order.
    E.g. ``p = 1.0`` and ``p = 2.0`` with yield the Manhattan and L2 distances, respectively.

        .. math:: D_{ij} = \\|A_i - B_j\\|_p

    Where :math:`A_{i}` and :math:`B_{j}` are representation vectors.
    D is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of descriptors - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of descriptors - shape (M, representation size).
    :type B: numpy array
    :param p: The norm order
    :type p: float

    :return: The distance matrix.
    :rtype: numpy array
    """

    if len(A.shape) != 2 or len(B.shape) != 2:
        raise ValueError("expected matrices of dimension=2")

    if B.shape[1] != A.shape[1]:
        raise ValueError("expected matrices containing vectors of same size")

    na = A.shape[0]
    nb = B.shape[0]

    D = np.empty((na, nb), order="F")

    if isinstance(p, int):
        if p == 2:
            fl2_distance(A, B, D)
        else:
            fp_distance_integer(A.T, B.T, D, p)

    elif isinstance(p, float):
        if p.is_integer():
            p = int(p)
            if p == 2:
                fl2_distance(A, B, D)
            else:
                fp_distance_integer(A.T, B.T, D, p)

        else:
            fp_distance_double(A.T, B.T, D, p)
    else:
        raise ValueError("expected exponent of integer or float type")

    return D



import ast
import csv
from copy import deepcopy

import numpy as np
import pytest
from scipy.linalg import lstsq

from qmllib.representations import (
    generate_fchl18,
    generate_fchl18_displaced,
    generate_fchl18_electric_field,
)
from qmllib.representations.fchl import (
    get_atomic_local_electric_field_gradient_kernels,
    get_atomic_local_gradient_kernels,
    get_atomic_local_kernels,
    get_gaussian_process_electric_field_kernels,
)
from qmllib.solvers import cho_solve

try:
    import rdkit
except ImportError:
    rdkit = None

try:
    import pybel
except ImportError:
    pybel = None

from conftest import ASSETS

needsrdkit = pytest.mark.skipif(rdkit is None, reason="Test requires RDKit installed")
needspybel = pytest.mark.skipif(pybel is None, reason="Test requires Pybel installed")


DEBYE_TO_EAA = 0.20819434
DEBYE_TO_AU = 0.393456
DEBYE_TO_EAA = 0.20819434
HARTREE_TO_KCAL_MOL = 627.509474
KCAL_MOL_TO_EV = 1.0 / 23.06035

HARTREE_TO_EV = HARTREE_TO_KCAL_MOL * KCAL_MOL_TO_EV
BOHR_TO_ANGS = 0.529177249

CUT_DISTANCE = 1e6
MAX_SIZE = 5

REP_ARGS = {"cut_distance": CUT_DISTANCE, "max_size": 5}

DX = 0.01
DF = 0.01
EF_SCALING = 0.01

SIGMAS = [0.64]

KERNEL_ARGS = {
    "kernel": "gaussian",
    "kernel_args": {
        "sigma": SIGMAS,
    },
    "cut_distance": CUT_DISTANCE,
    "alchemy": "off",
}


def parse_energy(filename):

    f = open(filename)
    lines = f.readlines()
    f.close()

    energy = dict()

    for line in lines:

        tokens = line.split()
        e = float(tokens[1])  # - -99.624524268 - -0.499821176)
        angle = ang2ang(float(tokens[0]))

        energy[angle] = e

    e = np.array([energy[key] for key in sorted(energy.keys())])
    offset = 0.0  # np.amin(e)

    for key in sorted(energy.keys()):
        energy[key] = energy[key] - offset  # * HARTREE_TO_KCAL_MOL * KCAL_MOL_TO_EV

    return energy


def ang2ang(angle):

    out = angle - 90.0

    if out < -180.0:
        out += 360

    return out


def parse_dipole(filename):

    f = open(filename)
    lines = f.readlines()
    f.close()

    dipole = dict()

    for line in lines:

        tokens = line.split()

        mu = np.array([float(tokens[-3]), float(tokens[-2]), float(tokens[-1])])
        angle = ang2ang(float(tokens[0]))

        dipole[angle] = mu  # * DEBYE_TO_EAA

    return dipole


def parse_csv(filename):

    X = []
    X_gradient = []
    X_dipole = []

    E = []
    G = []
    D = []

    with open(filename, "r") as csvfile:

        csvlines = csv.reader(csvfile, delimiter=";")

        for i, row in enumerate(csvlines):

            nuclear_charges = np.array(ast.literal_eval(row[6]), dtype=np.int32)

            # Gradients (from force in hartree/borh to gradients in eV/angstrom)
            gradient = np.array(ast.literal_eval(row[5])) * HARTREE_TO_EV / BOHR_TO_ANGS * -1

            # SCF energy (eV)
            energy = float(row[4]) * HARTREE_TO_EV

            # Dipole moment (Debye -> eV/Angs)
            dipole = np.array(ast.literal_eval(row[3])) * DEBYE_TO_EAA

            # Coordinates (Angstrom)
            coords = np.array(ast.literal_eval(row[2]))

            rep = generate_fchl18(nuclear_charges, coords, **REP_ARGS)
            rep_gradient = generate_fchl18_displaced(nuclear_charges, coords, dx=DX, **REP_ARGS)
            rep_dipole = generate_fchl18_electric_field(
                nuclear_charges, coords, fictitious_charges="Gasteiger", **REP_ARGS
            )

            E.append(energy)
            D.append(dipole)
            G.append(gradient)

            X.append(rep)
            X_gradient.append(rep_gradient)
            X_dipole.append(rep_dipole)

    X = np.array(X)
    X_gradient = np.array(X_gradient)
    X_dipole = np.array(X_dipole)

    E = np.array(E)
    G = np.array(G)
    D = np.array(D)

    return X, X_gradient, X_dipole, E, G, D


@pytest.mark.skip(reason="Missing test file")
def test_multiple_operators():

    X, X_gradient, X_dipole, E, G, D = parse_csv(ASSETS / "dichloromethane_mp2_test.csv")

    K = get_atomic_local_kernels(X, X, **KERNEL_ARGS)[0]
    K_gradient = get_atomic_local_gradient_kernels(X, X_gradient, dx=DX, **KERNEL_ARGS)[0]
    K_dipole = get_atomic_local_electric_field_gradient_kernels(
        X, X_dipole, df=DF, ef_scaling=EF_SCALING, **KERNEL_ARGS
    )[0]

    Xs, Xs_gradient, Xs_dipole, Es, Gs, Ds = parse_csv(ASSETS / "dichloromethane_mp2_train.csv")

    Ks = get_atomic_local_kernels(X, Xs, **KERNEL_ARGS)[0]
    Ks_gradient = get_atomic_local_gradient_kernels(X, Xs_gradient, dx=DX, **KERNEL_ARGS)[0]
    Ks_dipole = get_atomic_local_electric_field_gradient_kernels(
        X, Xs_dipole, df=DF, ef_scaling=EF_SCALING, **KERNEL_ARGS
    )[0]

    offset = E.mean()
    E -= offset
    Es -= offset

    Y = np.concatenate((E, G.flatten(), D.flatten()))
    C = np.concatenate((K.T, K_gradient.T, K_dipole.T))

    alpha, residuals, rank, sing = lstsq(C, Y, cond=-1, lapack_driver="gelsd")

    Et = np.dot(K.T, alpha)
    Gt = np.dot(K_gradient.T, alpha)
    Dt = np.dot(K_dipole.T, alpha)

    mae = np.mean(np.abs(Et - E)) / KCAL_MOL_TO_EV
    mae_gradient = np.mean(np.abs(Gt - G.flatten())) / KCAL_MOL_TO_EV * BOHR_TO_ANGS
    mae_dipole = np.mean(np.abs(Dt - D.flatten())) / DEBYE_TO_EAA

    print(mae)
    print(mae_gradient)
    print(mae_dipole)

    assert mae < 0.8, "Error in multiple operator training energy"
    assert mae_gradient < 0.1, "Error in multiple operator training energy"
    assert mae_dipole < 0.01, "Error in multiple operator training dipole"

    Ess = np.dot(Ks.T, alpha)
    Gss = np.dot(Ks_gradient.T, alpha)
    Dss = np.dot(Ks_dipole.T, alpha)

    mae = np.mean(np.abs(Ess - Es)) / KCAL_MOL_TO_EV
    mae_gradient = np.mean(np.abs(Gss - Gs.flatten())) / KCAL_MOL_TO_EV * BOHR_TO_ANGS
    mae_dipole = np.mean(np.abs(Dss - Ds.flatten())) / DEBYE_TO_EAA

    print(mae)
    print(mae_gradient)
    print(mae_dipole)

    assert mae < 0.8, "Error in multiple operator test energy"
    assert mae_gradient < 0.1, "Error in multiple operator test force"
    assert mae_dipole < 0.02, "Error in multiple operator test dipole"


def test_generate_representation():

    coords = np.array([[1.464, 0.707, 1.056], [0.878, 1.218, 0.498], [2.319, 1.126, 0.952]])

    nuclear_charges = np.array([8, 1, 1], dtype=np.int32)

    rep_ref = np.loadtxt(ASSETS / "fchl_ef_rep.txt").reshape((3, 6, 3))

    # Test with fictitious charges from a numpy array
    fic_charges1 = np.array([-0.41046649, 0.20523324, 0.20523324])

    rep1 = generate_fchl18_electric_field(
        nuclear_charges, coords, fictitious_charges=fic_charges1, max_size=3
    )

    assert np.allclose(rep1, rep_ref), "Error generating representation for electric fields"

    # Test with fictitious charges from a list
    fic_charges2 = [-0.41046649, 0.20523324, 0.20523324]

    rep2 = generate_fchl18_electric_field(
        nuclear_charges, coords, fictitious_charges=fic_charges2, max_size=3
    )

    assert np.allclose(rep2, rep_ref), "Error generating representation for electric fields"


@needspybel()
def test_generate_representation_pybel():
    # TODO Generate Gastiger charges with pybel

    # Test with fictitious charges from Open Babel (Gasteiger).
    # Skip test if there is no pybel/openbabel
    # try:
    #     pass
    # except ImportError:
    #     return

    # rep3 = generate_representation_electric_field(
    # nuclear_charges,
    # coords,
    # fictitious_charges="Gasteiger", max_size=3
    # )

    # assert np.allclose(
    #     rep3, rep_ref
    # ), "Error assigning partial charges: Check Openbabel/Pybel installation"

    raise NotImplementedError()


@needsrdkit()
def test_generate_representation_rdkit():
    # TODO Generate Gastiger charges with rdkit
    raise NotImplementedError()


@pytest.mark.skip(reason="Missing test file")
def test_gaussian_process():

    X, X_gradient, X_dipole, E, G, D = parse_csv(ASSETS / "dichloromethane_mp2_test.csv")

    K = get_gaussian_process_electric_field_kernels(X_dipole, X_dipole, **KERNEL_ARGS)[0]

    Xs, Xs_gradient, Xs_dipole, Es, Gs, Ds = parse_csv(ASSETS / "dichloromethane_mp2_train.csv")

    Ks = get_gaussian_process_electric_field_kernels(X_dipole, Xs_dipole, **KERNEL_ARGS)[0]

    offset = E.mean()
    E -= offset
    Es -= offset

    Y = np.concatenate((E, D.flatten()))
    # UNUSED Ys = np.concatenate((Es, Ds.flatten()))

    C = deepcopy(K)
    C[np.diag_indices_from(C)] += 1e-9
    alpha = cho_solve(C, Y)

    Yt = np.dot(K.T, alpha)
    Yss = np.dot(Ks.T, alpha)

    n = len(E)

    mae = np.mean(np.abs(Yt[:n] - E)) / KCAL_MOL_TO_EV
    mae_dipole = np.mean(np.abs(Yt[n:] - D.flatten())) / DEBYE_TO_EAA

    print(mae)
    print(mae_dipole)

    assert mae < 0.002, "Error in multiple operator training energy"
    assert mae_dipole < 0.001, "Error in multiple operator training dipole"

    mae = np.mean(np.abs(Yss[:n] - Es)) / KCAL_MOL_TO_EV
    mae_dipole = np.mean(np.abs(Yss[n:] - Ds.flatten())) / DEBYE_TO_EAA

    print(mae)
    print(mae_dipole)

    assert mae < 0.3, "Error in multiple operator test energy"
    assert mae_dipole < 0.02, "Error in multiple operator test dipole"


@pytest.mark.skip(reason="Missing test files")
def test_gaussian_process_field_dependent():

    dipole = parse_dipole(ASSETS / "hf_dipole.txt")
    energy = parse_energy(ASSETS / "hf_energy.txt")

    # Get energies, dipole moments and angles from datafiles
    # UNUSED e = np.array([energy[key] for key in sorted(energy.keys())])
    # UNUSED mu = np.array([dipole[key] for key in sorted(dipole.keys())])
    # UNUSED a = np.array([key for key in sorted(dipole.keys())])

    # Generate dummy coordinates
    coordinates = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]])
    nuclear_charges = np.array([9, 1], dtype=np.int32)

    # Which angle to put in the training set
    train_angles = np.array([135])

    reps = []
    fields = []
    eY = []
    dY = []

    # Make training set
    for ang in train_angles:

        ang_rad = ang / 180.0 * np.pi

        field = np.array([np.cos(ang_rad), np.sin(ang_rad), 0.0]) * 0.001
        rep = generate_fchl18_electric_field(
            nuclear_charges, coordinates, max_size=2, neighbors=2, cut_distance=1e6
        )
        fields.append(field)
        reps.append(rep)

        eY.append(energy[ang])

        dY.append(dipole[ang])

    F = [fields, fields]
    X = np.array(reps)
    eY = np.array(eY)
    dY = np.concatenate(dY)

    eYs = []
    dYs = []

    reps = []
    fields_test = []

    # Make test set
    test_angles = range(-180, 180, 20)
    for ang in test_angles:

        ang_rad = ang / 180.0 * np.pi

        field = np.array([np.cos(ang_rad), np.sin(ang_rad), 0.0]) * 0.001

        rep = generate_fchl18_electric_field(
            nuclear_charges, coordinates, max_size=2, neighbors=2, cut_distance=1e6
        )

        fields_test.append(field)
        reps.append(rep)

        eYs.append(energy[ang])
        dYs.append(dipole[ang])

    Fs = [fields, fields_test]
    Xs = np.array(reps)
    eYs = np.array(eYs)
    dYs = np.concatenate(dYs)

    Y = np.concatenate((eY, dY))

    sigmas = [2.5]
    kernel_args = {
        "kernel": "gaussian",
        "kernel_args": {
            "sigma": sigmas,
        },
        "cut_distance": 1e6,
        "alchemy": "off",
    }

    # Get Gaussian Process kernels for training and test
    K = get_gaussian_process_electric_field_kernels(X, X, fields=F, **kernel_args)[0]
    Ks = get_gaussian_process_electric_field_kernels(X, Xs, fields=Fs, **kernel_args)[0]

    n_train = len(eY)
    n_test = len(eYs)

    # Solve
    C = deepcopy(K)
    C[np.diag_indices_from(C)] += 1e-14
    alpha = cho_solve(C, Y)

    Cs = deepcopy(Ks)

    Yt = np.dot(K.T, alpha)

    eYt = Yt[:n_train]
    dYt = Yt[n_train:]

    # get predictions
    Yss = np.dot(Cs.T, alpha)

    eYss = Yss[:n_test]
    dYss = Yss[n_test:]

    dtmae = np.mean(np.abs(dYt - dY))
    etmae = np.mean(np.abs(eYt - eY))

    dmae = np.mean(np.abs(dYs - dYss))
    emae = np.mean(np.abs(eYs - eYss))

    # print("%7.2f    %10.4f  %14.4f  %14.4f  %14.4f" % (sigmas[0], dmae, emae,dtmae,etmae))

    assert dmae < 0.005, "Error in test dipole"
    assert dtmae < 0.002, "Error in training dipole"
    assert emae < 0.0001, "Error in test energy"
    assert etmae < 0.0001, "Error in training energy"



import numpy as np
from conftest import ASSETS, get_energies, shuffle_arrays

from qmllib.kernels import laplacian_kernel
from qmllib.representations.bob import get_asize
from qmllib.representations.representations import generate_bob
from qmllib.solvers import cho_solve
from qmllib.utils.xyz_format import read_xyz


def test_krr_bob():

    # Parse file containing PBE0/def2-TZVP heats of formation and xyz filenames
    n_points = 1000
    data = get_energies(ASSETS / "hof_qm7.txt")
    filenames = sorted(data.keys())[:n_points]

    molecules = []
    representations = []
    properties = []

    for filename in filenames:
        coord, atoms = read_xyz((ASSETS / "qm7" / filename).with_suffix(".xyz"))
        molecules.append((coord, atoms))
        properties.append(data[filename])

    size = max(atoms.size for _, atoms in molecules) + 1
    # example asize={"O": 3, "C": 7, "N": 3, "H": 16, "S": 1},
    asize = get_asize([atoms for _, atoms in molecules], 1)

    atomtypes = []
    for _, atoms in molecules:
        atomtypes.extend(atoms)
    atomtypes = np.unique(atomtypes)

    for coord, atoms in molecules:
        # representation = generate_bob(atoms, coord, atomtypes, )
        rep = generate_bob(atoms, coord, atomtypes, size=size, asize=asize)
        representations.append(rep)

    representations = np.array(representations)
    properties = np.array(properties)
    shuffle_arrays(properties, representations, seed=666)

    # Make training and test sets
    n_test = 300
    n_train = 700
    train_indices = list(range(n_train))
    test_indices = list(range(n_train, n_train + n_test))

    # List of representations and properties
    test_representations = representations[test_indices]
    train_representations = representations[train_indices]
    test_properties = properties[test_indices]
    train_properties = properties[train_indices]

    # Set hyper-parameters
    sigma = 26214.40
    llambda = 1e-10

    # Generate training Kernel
    K = laplacian_kernel(train_representations, train_representations, sigma)

    # Solve alpha
    K[np.diag_indices_from(K)] += llambda
    alpha = cho_solve(K, train_properties)

    # Calculate prediction kernel
    Ks = laplacian_kernel(train_representations, test_representations, sigma)
    predicted_properties = np.dot(Ks.transpose(), alpha)

    mae = np.mean(np.abs(test_properties - predicted_properties))
    print(mae)
    assert mae < 2.6, "ERROR: Too high MAE!"



import contextlib
import ctypes


def mkl_set_num_threads(cores: int) -> None:

    try:
        mkl_rt = ctypes.CDLL("libmkl_rt.so")
        mkl_rt.mkl_set_num_threads(ctypes.byref(ctypes.c_int(cores)))
    except OSError:
        pass


def mkl_get_num_threads():

    try:
        mkl_rt = ctypes.CDLL("libmkl_rt.so")
        mkl_num_threads = mkl_rt.mkl_get_max_threads()
        return mkl_num_threads

    except OSError:

        return None


def mkl_reset_num_threads():

    try:
        mkl_rt = ctypes.CDLL("libmkl_rt.so")
        mkl_num_threads = mkl_rt.mkl_get_max_threads()
        mkl_rt.mkl_set_num_threads(ctypes.byref(ctypes.c_int(mkl_num_threads)))

    except OSError:
        pass


@contextlib.contextmanager
def modified_environ(*remove, **update):
    """Temporarily updates the environment dictionary in-place."""
    # TODO Use context manager for thread cores changes
    raise NotImplementedError()



"""
Bag of bonds utils functions
"""

from typing import List

import numpy as np
from numpy import ndarray

from qmllib.constants.periodic_table import ELEMENT_NAME


def get_natypes(nuclear_charges: np.ndarray) -> dict[str, int]:
    """Get number of atom types, from a list of molecules"""

    keys, counts = np.unique(nuclear_charges, return_counts=True)

    # natypes = dict([(key, len(value)) for key,value in self.atomtype_indices.items()])

    # natypes = dict([ (key, count) for key, count in zip(keys, counts)])

    keys_name = [ELEMENT_NAME[key] for key in keys]

    natypes = dict([(key, count) for key, count in zip(keys_name, counts)])

    return natypes


def get_asize(list_nuclear_charges: List[ndarray], pad: int) -> dict[str, int]:
    """

    example:
        asize = {"O":3, "C":7, "N":3, "H":16, "S":1}
    """

    asize: dict[str, int] = dict()

    for nuclear_charges in list_nuclear_charges:
        natypes = get_natypes(nuclear_charges)
        for key, value in natypes.items():
            try:
                asize[key] = max(asize[key], value + pad)
            except KeyError:
                asize[key] = value + pad
    return asize



# flake8: noqa

import ast
import csv
from copy import deepcopy

import numpy as np
import pytest
import scipy
import scipy.stats
from conftest import ASSETS
from scipy.linalg import lstsq

from qmllib.representations import (
    generate_fchl18,
    generate_fchl18_displaced,
    generate_fchl18_displaced_5point,
)
from qmllib.representations.fchl import (
    get_atomic_local_gradient_5point_kernels,
    get_atomic_local_gradient_kernels,
    get_atomic_local_kernels,
    get_force_alphas,
    get_gaussian_process_kernels,
    get_local_gradient_kernels,
    get_local_hessian_kernels,
    get_local_kernels,
    get_local_symmetric_hessian_kernels,
    get_local_symmetric_kernels,
)
from qmllib.solvers import cho_solve

FORCE_KEY = "forces"
ENERGY_KEY = "om2_energy"
CSV_FILE = ASSETS / "amons_small.csv"
SIGMAS = [0.64]

TRAINING = 13
TEST = 7

DX = 0.005
CUT_DISTANCE = 1e6
KERNEL_ARGS = {
    "verbose": False,
    "cut_distance": CUT_DISTANCE,
    "kernel": "gaussian",
    "kernel_args": {
        "sigma": SIGMAS,
    },
}

LLAMBDA_ENERGY = 1e-7
LLAMBDA_FORCE = 1e-7


pytest.skip(allow_module_level=True, reason="Test is broken")


def mae(a, b):

    return np.mean(np.abs(a.flatten() - b.flatten()))


def csv_to_molecular_reps(csv_filename, force_key="orca_forces", energy_key="orca_energy"):

    np.random.seed(667)

    x = []
    f = []
    e = []
    distance = []

    disp_x = []
    disp_x5 = []

    max_atoms = 5

    with open(csv_filename, "r") as csvfile:

        df = csv.reader(csvfile, delimiter=";", quotechar="#")

        for row in df:

            coordinates = np.array(ast.literal_eval(row[2]))
            nuclear_charges = ast.literal_eval(row[5])
            atomtypes = ast.literal_eval(row[1])
            force = np.array(ast.literal_eval(row[3]))
            energy = float(row[6])

            rep = generate_fchl18(
                coordinates, nuclear_charges, max_size=max_atoms, cut_distance=CUT_DISTANCE
            )

            disp_rep = generate_fchl18_displaced(
                coordinates, nuclear_charges, max_size=max_atoms, cut_distance=CUT_DISTANCE, dx=DX
            )

            disp_rep5 = generate_fchl18_displaced_5point(
                coordinates, nuclear_charges, max_size=max_atoms, cut_distance=CUT_DISTANCE, dx=DX
            )

            x.append(rep)
            f.append(force)
            e.append(energy)

            disp_x.append(disp_rep)
            disp_x5.append(disp_rep5)

    return np.array(x), f, e, np.array(disp_x), np.array(disp_x5)


def test_gaussian_process_derivative():

    Xall, Fall, Eall, dXall, dXall5 = csv_to_molecular_reps(
        CSV_FILE, force_key=FORCE_KEY, energy_key=ENERGY_KEY
    )

    Eall = np.array(Eall)
    Fall = np.array(Fall)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    F = Fall[:TRAINING]
    E = Eall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]
    Fs = Fall[-TEST:]
    Es = Eall[-TEST:]

    K = get_gaussian_process_kernels(X, dX, dx=DX, **KERNEL_ARGS)
    Kt = K[:, TRAINING:, TRAINING:]
    Kt_local = K[:, :TRAINING, :TRAINING]
    Kt_energy = K[:, :TRAINING, TRAINING:]

    Kt_grad2 = get_local_gradient_kernels(X, dX, dx=DX, **KERNEL_ARGS)

    Ks = get_local_hessian_kernels(dX, dXs, dx=DX, **KERNEL_ARGS)
    Ks_energy = get_local_gradient_kernels(X, dXs, dx=DX, **KERNEL_ARGS)

    Ks_energy2 = get_local_gradient_kernels(Xs, dX, dx=DX, **KERNEL_ARGS)
    Ks_local = get_local_kernels(X, Xs, **KERNEL_ARGS)

    F = np.concatenate(F)
    Fs = np.concatenate(Fs)

    Y = np.array(F.flatten())
    Y = np.concatenate((E, Y))

    for i, sigma in enumerate(SIGMAS):

        C = deepcopy(K[i])

        for j in range(TRAINING):
            C[j, j] += LLAMBDA_ENERGY

        for j in range(TRAINING, K.shape[2]):
            C[j, j] += LLAMBDA_FORCE

        alpha = cho_solve(C, Y)
        beta = alpha[:TRAINING]
        gamma = alpha[TRAINING:]

        Fss = np.dot(np.transpose(Ks[i]), gamma) + np.dot(np.transpose(Ks_energy[i]), beta)
        Ft = np.dot(np.transpose(Kt[i]), gamma) + np.dot(np.transpose(Kt_energy[i]), beta)

        Ess = np.dot(Ks_energy2[i], gamma) + np.dot(Ks_local[i].T, beta)
        Et = np.dot(Kt_energy[i], gamma) + np.dot(Kt_local[i].T, beta)

        assert mae(Ess, Es) < 0.1, "Error in Gaussian Process test energy"
        assert mae(Et, E) < 0.001, "Error in Gaussian Process training energy"

        assert mae(Fss, Fs) < 1.0, "Error in Gaussian Process test force"
        assert mae(Ft, F) < 0.001, "Error in Gaussian Process training force"


def test_gdml_derivative():

    Xall, Fall, Eall, dXall, dXall5 = csv_to_molecular_reps(
        CSV_FILE, force_key=FORCE_KEY, energy_key=ENERGY_KEY
    )

    Eall = np.array(Eall)
    Fall = np.array(Fall)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    F = Fall[:TRAINING]
    E = Eall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]
    Fs = Fall[-TEST:]
    Es = Eall[-TEST:]

    K = get_local_symmetric_hessian_kernels(dX, dx=DX, **KERNEL_ARGS)
    Ks = get_local_hessian_kernels(dXs, dX, dx=DX, **KERNEL_ARGS)

    Kt_energy = get_local_gradient_kernels(X, dX, dx=DX, **KERNEL_ARGS)
    Ks_energy = get_local_gradient_kernels(Xs, dX, dx=DX, **KERNEL_ARGS)

    F = np.concatenate(F)
    Fs = np.concatenate(Fs)

    Y = np.array(F.flatten())
    # Y = np.concatenate((E, Y))

    for i, sigma in enumerate(SIGMAS):

        C = deepcopy(K[i])
        for j in range(K.shape[2]):
            C[j, j] += LLAMBDA_FORCE

        alpha = cho_solve(C, Y)
        Fss = np.dot(Ks[i], alpha)
        Ft = np.dot(K[i], alpha)

        Ess = np.dot(Ks_energy[i], alpha)
        Et = np.dot(Kt_energy[i], alpha)

        slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(
            E.flatten(), Et.flatten()
        )

        Ess -= intercept
        Et -= intercept

        # This test will only work for molecules of same type
        # assert mae(Ess, Es) < 0.1, "Error in Gaussian Process test energy"
        # assert mae(Et, E) < 0.001, "Error in Gaussian Process training energy"

        assert mae(Fss, Fs) < 1.0, "Error in GDML test force"
        assert mae(Ft, F) < 0.001, "Error in GDML training force"


def test_normal_equation_derivative():

    Xall, Fall, Eall, dXall, dXall5 = csv_to_molecular_reps(
        CSV_FILE, force_key=FORCE_KEY, energy_key=ENERGY_KEY
    )

    Eall = np.array(Eall)
    Fall = np.array(Fall)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    dX5 = dXall5[:TRAINING]
    F = Fall[:TRAINING]
    E = Eall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]
    dXs5 = dXall5[-TEST:]
    Fs = Fall[-TEST:]
    Es = Eall[-TEST:]

    Ftrain = np.concatenate(F)
    Etrain = np.array(E)
    alphas = get_force_alphas(
        X, dX, Ftrain, energy=Etrain, dx=DX, regularization=LLAMBDA_FORCE, **KERNEL_ARGS
    )

    Kt_force = get_atomic_local_gradient_kernels(X, dX, dx=DX, **KERNEL_ARGS)
    Ks_force = get_atomic_local_gradient_kernels(X, dXs, dx=DX, **KERNEL_ARGS)

    Kt_force5 = get_atomic_local_gradient_5point_kernels(X, dX5, dx=DX, **KERNEL_ARGS)
    Ks_force5 = get_atomic_local_gradient_5point_kernels(X, dXs5, dx=DX, **KERNEL_ARGS)

    Kt_energy = get_atomic_local_kernels(X, X, **KERNEL_ARGS)
    Ks_energy = get_atomic_local_kernels(X, Xs, **KERNEL_ARGS)

    F = np.concatenate(F)
    Fs = np.concatenate(Fs)
    Y = np.array(F.flatten())

    for i, sigma in enumerate(SIGMAS):

        Ft = np.zeros((Kt_force[i, :, :].shape[1] // 3, 3))
        Fss = np.zeros((Ks_force[i, :, :].shape[1] // 3, 3))

        Ft5 = np.zeros((Kt_force5[i, :, :].shape[1] // 3, 3))
        Fss5 = np.zeros((Ks_force5[i, :, :].shape[1] // 3, 3))

        for xyz in range(3):

            Ft[:, xyz] = np.dot(Kt_force[i, :, xyz::3].T, alphas[i])
            Fss[:, xyz] = np.dot(Ks_force[i, :, xyz::3].T, alphas[i])

            Ft5[:, xyz] = np.dot(Kt_force5[i, :, xyz::3].T, alphas[i])
            Fss5[:, xyz] = np.dot(Ks_force5[i, :, xyz::3].T, alphas[i])

        Ess = np.dot(Ks_energy[i].T, alphas[i])
        Et = np.dot(Kt_energy[i].T, alphas[i])

        assert mae(Ess, Es) < 0.3, "Error in normal equation test energy"
        assert mae(Et, E) < 0.08, "Error in normal equation training energy"

        assert mae(Fss, Fs) < 3.2, "Error in  normal equation test force"
        assert mae(Ft, F) < 0.5, "Error in  normal equation training force"

        assert mae(Fss5, Fs) < 3.2, "Error in normal equation 5-point test force"
        assert mae(Ft5, F) < 0.5, "Error in normal equation 5-point training force"

        assert mae(Fss5, Fss) < 0.01, "Error in normal equation 5-point or 2-point test force"
        assert mae(Ft5, Ft) < 0.01, "Error in normal equation 5-point or 2-point training force"


def test_operator_derivative():

    Xall, Fall, Eall, dXall, dXall5 = csv_to_molecular_reps(
        CSV_FILE, force_key=FORCE_KEY, energy_key=ENERGY_KEY
    )

    Eall = np.array(Eall)
    Fall = np.array(Fall)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    dX5 = dXall5[:TRAINING]
    F = Fall[:TRAINING]
    E = Eall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]
    dXs5 = dXall5[-TEST:]
    Fs = Fall[-TEST:]
    Es = Eall[-TEST:]

    Ftrain = np.concatenate(F)
    Etrain = np.array(E)

    Kt_energy = get_atomic_local_kernels(X, X, **KERNEL_ARGS)
    Ks_energy = get_atomic_local_kernels(X, Xs, **KERNEL_ARGS)

    Kt_force = get_atomic_local_gradient_kernels(X, dX, dx=DX, **KERNEL_ARGS)
    Ks_force = get_atomic_local_gradient_kernels(X, dXs, dx=DX, **KERNEL_ARGS)

    F = np.concatenate(F)
    Fs = np.concatenate(Fs)
    Y = np.array(F.flatten())

    for i, sigma in enumerate(SIGMAS):

        Y = np.concatenate((E, F.flatten()))

        C = np.concatenate((Kt_energy[i].T, Kt_force[i].T))

        alphas, residuals, singular_values, rank = lstsq(C, Y, cond=1e-9, lapack_driver="gelsd")

        Ess = np.dot(Ks_energy[i].T, alphas)
        Et = np.dot(Kt_energy[i].T, alphas)

        Fss = np.dot(Ks_force[i].T, alphas)
        Ft = np.dot(Kt_force[i].T, alphas)

        assert mae(Ess, Es) < 0.08, "Error in operator test energy"
        assert mae(Et, E) < 0.04, "Error in  operator training energy"

        assert mae(Fss, Fs.flatten()) < 1.1, "Error in  operator test force"
        assert mae(Ft, F.flatten()) < 0.1, "Error in  operator training force"


def test_krr_derivative():

    Xall, Fall, Eall, dXall, dXall5 = csv_to_molecular_reps(
        CSV_FILE, force_key=FORCE_KEY, energy_key=ENERGY_KEY
    )

    Eall = np.array(Eall)
    Fall = np.array(Fall)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    F = Fall[:TRAINING]
    E = Eall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]
    Fs = Fall[-TEST:]
    Es = Eall[-TEST:]

    K = get_local_symmetric_kernels(X, **KERNEL_ARGS)
    Ks = get_local_kernels(Xs, X, **KERNEL_ARGS)

    Kt_force = get_local_gradient_kernels(X, dX, dx=DX, **KERNEL_ARGS)
    Ks_force = get_local_gradient_kernels(X, dXs, dx=DX, **KERNEL_ARGS)

    F = np.concatenate(F)
    Fs = np.concatenate(Fs)

    Y = np.array(E)

    for i, sigma in enumerate(SIGMAS):

        C = deepcopy(K[i])
        for j in range(K.shape[2]):
            C[j, j] += LLAMBDA_ENERGY

        alpha = cho_solve(C, Y)

        Fss = np.dot(Ks_force[i].T, alpha)
        Ft = np.dot(Kt_force[i].T, alpha)

        Ess = np.dot(Ks[i], alpha)
        Et = np.dot(K[i], alpha)

        slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(
            E.flatten(), Et.flatten()
        )

        assert mae(Ess, Es) < 0.7, "Error in KRR test energy"
        assert mae(Et, E) < 0.02, "Error in KRR training energy"

        assert mae(Fss, Fs) < 5.6, "Error in KRR test force"
        assert mae(Ft, F) < 4.3, "Error in KRR training force"


if __name__ == "__main__":

    test_gaussian_process_derivative()
    test_gdml_derivative()
    test_normal_equation_derivative()
    test_operator_derivative()
    test_krr_derivative()



import numpy as np
from conftest import ASSETS, get_energies, shuffle_arrays

from qmllib.kernels import get_local_kernel, get_local_symmetric_kernel
from qmllib.representations import generate_fchl19
from qmllib.solvers import cho_solve
from qmllib.utils.xyz_format import read_xyz

np.set_printoptions(linewidth=666)


def test_energy():

    # Read the heat-of-formation energies
    data = get_energies(ASSETS / "hof_qm7.txt")

    # Generate a list
    all_representations = []
    all_properties = []
    all_atoms = []

    for xyz_file in sorted(data.keys())[:1000]:

        filename = ASSETS / "qm7" / xyz_file
        coord, atoms = read_xyz(filename)

        # Associate a property (heat of formation) with the object
        all_properties.append(data[xyz_file])

        representation = generate_fchl19(atoms, coord, gradients=False, pad=27)

        all_representations.append(representation)
        all_atoms.append(atoms)

    # Convert to arrays
    all_representations = np.array(all_representations)
    all_properties = np.array(all_properties)
    # all_atoms = np.array(all_atoms)

    shuffle_arrays(all_representations, all_atoms, all_properties, seed=666)

    # Make training and test sets
    n_test = 99
    n_train = 101

    train_indices = list(range(n_train))
    test_indices = list(range(n_train, n_train + n_test))

    # List of representations
    test_representations = all_representations[test_indices]
    train_representations = all_representations[train_indices]
    test_atoms = [all_atoms[i] for i in test_indices]
    train_atoms = [all_atoms[i] for i in train_indices]
    test_properties = all_properties[test_indices]
    train_properties = all_properties[train_indices]

    # Set hyper-parameters
    sigma = 3.0
    llambda = 1e-10

    kernel = get_local_symmetric_kernel(train_representations, train_atoms, sigma)

    # Solve alpha
    alpha = cho_solve(kernel, train_properties, l2reg=llambda)

    # Calculate test kernel
    # test_kernel = get_local_kernel(train_representations, test_representations, train_atoms, test_atoms, sigma)

    # Calculate test prediction kernel
    prediction_kernel = get_local_kernel(
        train_representations, test_representations, train_atoms, test_atoms, sigma
    )
    prediction_properties = np.dot(prediction_kernel, alpha)

    mae = np.mean(np.abs(test_properties - prediction_properties))
    # assert mae < 4.0, "ERROR: Too high MAE!"
    assert mae < 4.9, "ERROR: Too high MAE!"



from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from numpy import ndarray
from scipy.special import binom, factorial

from .ffchl_module import ffchl_kernel_types as kt


def get_gaussian_parameters(
    tags: Optional[Dict[str, List[float]]]
) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "sigma": [2.5],
        }

    parameters = np.array(tags["sigma"])

    for i in range(len(parameters)):
        parameters[i] = -0.5 / (parameters[i]) ** 2

    np.resize(parameters, (1, len(tags["sigma"])))

    n_kernels = len(tags["sigma"])

    return kt.gaussian, parameters, n_kernels


def get_linear_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "c": [0.0],
        }

    parameters = np.array(tags["c"])

    np.resize(parameters, (1, len(tags["c"])))

    n_kernels = len(tags["c"])

    return kt.linear, parameters, n_kernels


def get_polynomial_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {"alpha": [1.0], "c": [0.0], "d": [1.0]}

    parameters = np.array([tags["alpha"], tags["c"], tags["d"]]).T
    if not (len(tags["alpha"]) == len(tags["c"]) or len(tags["alpha"]) == len(tags["d"])):
        raise ValueError("Unexpected parameter dimensions")

    n_kernels = len(tags["alpha"])
    return kt.polynomial, parameters, n_kernels


def get_sigmoid_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "alpha": [1.0],
            "c": [0.0],
        }

    parameters = np.array(
        [
            tags["alpha"],
            tags["c"],
        ]
    ).T
    if not len(tags["alpha"]) == len(tags["c"]):
        raise ValueError("Unexpected parameter dimensions")

    n_kernels = len(tags["alpha"])

    return kt.sigmoid, parameters, n_kernels


def get_multiquadratic_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "c": [0.0],
        }

    parameters = np.array(
        [
            tags["c"],
        ]
    ).T

    np.resize(parameters, (1, len(tags["c"])))
    n_kernels = len(tags["c"])

    return kt.multiquadratic, parameters, n_kernels


def get_inverse_multiquadratic_parameters(
    tags: Dict[str, List[float]]
) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "c": [0.0],
        }

    parameters = np.array(
        [
            tags["c"],
        ]
    ).T

    np.resize(parameters, (1, len(tags["c"])))
    n_kernels = len(tags["c"])

    return kt.inv_multiquadratic, parameters, n_kernels


def get_bessel_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {"sigma": [1.0], "v": [1.0], "n": [1.0]}

    parameters = np.array([tags["sigma"], tags["v"], tags["n"]]).T
    if not (len(tags["sigma"]) == len(tags["v"]) or len(tags["sigma"]) == len(tags["n"])):
        raise ValueError("Unexpected parameter dimensions")

    n_kernels = len(tags["sigma"])

    return kt.bessel, parameters, n_kernels


def get_l2_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "alpha": [1.0],
            "c": [0.0],
        }

    parameters = np.array(
        [
            tags["alpha"],
            tags["c"],
        ]
    ).T
    if not len(tags["alpha"]) == len(tags["c"]):
        raise ValueError("Unexpected parameter dimensions")
    n_kernels = len(tags["alpha"])

    return kt.l2, parameters, n_kernels


def get_matern_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "sigma": [10.0],
            "n": [2.0],
        }

    if not len(tags["sigma"]) == len(tags["n"]):
        raise ValueError("Unexpected parameter dimensions")

    n_kernels = len(tags["sigma"])

    n_max = int(max(tags["n"])) + 1

    parameters = np.zeros((2 + n_max, n_kernels))

    for i in range(n_kernels):
        parameters[0, i] = tags["sigma"][i]
        parameters[1, i] = tags["n"][i]

        n = int(tags["n"][i])
        for k in range(0, n + 1):
            parameters[2 + k, i] = float(factorial(n + k) * binom(n, k)) / factorial(2 * n)

    parameters = parameters.T

    return kt.matern, parameters, n_kernels


def get_cauchy_parameters(tags: Dict[str, List[float]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "sigma": [1.0],
        }

    parameters = np.array(
        [
            tags["sigma"],
        ]
    ).T

    np.resize(parameters, (1, len(tags["sigma"])))
    n_kernels = len(tags["sigma"])

    return kt.cauchy, parameters, n_kernels


def get_polynomial2_parameters(tags: Dict[str, List[List[float]]]) -> Tuple[ndarray, ndarray, int]:

    if tags is None:
        tags = {
            "coeff": [[1.0, 1.0, 1.0]],
        }

    parameters = np.zeros((10, len(tags["coeff"])))

    for i, c in enumerate(tags["coeff"]):
        for j, v in enumerate(c):
            parameters[j, i] = v

    n_kernels = len(tags["coeff"])
    parameters = parameters.T
    return kt.polynomial2, parameters, n_kernels


def get_kernel_parameters(
    name: str, tags: Optional[Union[Dict[str, List[List[float]]], Dict[str, List[float]]]]
) -> Tuple[ndarray, ndarray, int]:

    parameters = None
    idx = kt.gaussian
    n_kernels = 1

    if name == "gaussian":
        idx, parameters, n_kernels = get_gaussian_parameters(tags)

    elif name == "linear":
        idx, parameters, n_kernels = get_linear_parameters(tags)

    elif name == "polynomial":
        idx, parameters, n_kernels = get_polynomial_parameters(tags)

    elif name == "sigmoid":
        idx, parameters, n_kernels = get_sigmoid_parameters(tags)

    elif name == "multiquadratic":
        idx, parameters, n_kernels = get_multiquadratic_parameters(tags)

    elif name == "inverse-multiquadratic":
        idx, parameters, n_kernels = get_inverse_multiquadratic_parameters(tags)

    elif name == "bessel":
        idx, parameters, n_kernels = get_bessel_parameters(tags)

    elif name == "l2":
        idx, parameters, n_kernels = get_l2_parameters(tags)

    elif name == "matern":
        idx, parameters, n_kernels = get_matern_parameters(tags)

    elif name == "cauchy":
        idx, parameters, n_kernels = get_cauchy_parameters(tags)

    elif name == "polynomial2":
        idx, parameters, n_kernels = get_polynomial2_parameters(tags)

    else:
        raise ValueError(f"Unsupported kernel specification '{name}")

    return idx, parameters, n_kernels



import numpy as np
from conftest import ASSETS, get_energies

from qmllib.representations.arad import (
    generate_arad,
    get_atomic_kernels_arad,
    get_atomic_symmetric_kernels_arad,
    get_global_kernels_arad,
    get_global_symmetric_kernels_arad,
    get_local_kernels_arad,
    get_local_symmetric_kernels_arad,
)
from qmllib.utils.xyz_format import read_xyz


def test_arad():

    # Parse file containing PBE0/def2-TZVP heats of formation and xyz filenames
    n_points = 10
    data = get_energies(ASSETS / "hof_qm7.txt")
    filenames = sorted(data.keys())[:n_points]

    molecules = []
    representations = []
    properties = []

    for filename in filenames:
        coord, atoms = read_xyz((ASSETS / "qm7" / filename).with_suffix(".xyz"))
        molecules.append((coord, atoms))
        properties.append(data[filename])

    for coord, atoms in molecules:
        rep = generate_arad(atoms, coord)
        representations.append(rep)

    representations = np.array(representations)
    properties = np.array(properties)

    # for xyz_file in sorted(data.keys())[:10]:

    #     # Initialize the qmllib.data.Compound() objects
    #     mol = qmllib.Compound(xyz=test_dir + "/qm7/" + xyz_file)

    #     # Associate a property (heat of formation) with the object
    #     mol.properties = data[xyz_file]

    #     # This is a Molecular Coulomb matrix sorted by row norm

    #     representation = generate_arad_representation(mol.coordinates, mol.nuclear_charges)

    #     mols.append(mol)

    sigmas = [25.0]

    # X1 = np.array([mol.representation for mol in mols])

    K_local_asymm = get_local_kernels_arad(representations, representations, sigmas)
    K_local_symm = get_local_symmetric_kernels_arad(representations, sigmas)

    assert np.allclose(K_local_symm, K_local_asymm), "Symmetry error in local kernels"
    assert np.invert(
        np.all(np.isnan(K_local_asymm))
    ), "ERROR: ARAD local symmetric kernel contains NaN"

    K_global_asymm = get_global_kernels_arad(representations, representations, sigmas)
    K_global_symm = get_global_symmetric_kernels_arad(representations, sigmas)

    assert np.allclose(K_global_symm, K_global_asymm), "Symmetry error in global kernels"
    assert np.invert(
        np.all(np.isnan(K_global_asymm))
    ), "ERROR: ARAD global symmetric kernel contains NaN"

    molid = 5
    coordinates, atoms = molecules[molid]
    natoms = len(atoms)
    X1 = generate_arad(atoms, coordinates, size=natoms)
    XA = X1[:natoms]

    K_atomic_asymm = get_atomic_kernels_arad(XA, XA, sigmas)
    K_atomic_symm = get_atomic_symmetric_kernels_arad(XA, sigmas)

    assert np.allclose(K_atomic_symm, K_atomic_asymm), "Symmetry error in atomic kernels"
    assert np.invert(
        np.all(np.isnan(K_atomic_asymm))
    ), "ERROR: ARAD atomic symmetric kernel contains NaN"

    K_atomic_asymm = get_atomic_kernels_arad(XA, XA, sigmas)
